"""this is a prototype ensemble smoother based on the LM-EnRML
algorithm of Chen and Oliver 2013.  It requires the pest++ "sweep" utility
to propagate the ensemble forward.
"""
from __future__ import print_function, division
import os
from datetime import datetime
import shutil
import threading
import time
import warnings

import numpy as np
import pandas as pd
import pyemu
from pyemu.en import ParameterEnsemble,ObservationEnsemble
from pyemu.mat import Cov,Matrix

from pyemu.pst import Pst
from ..logger import Logger
from .ensemble_method import EnsembleMethod

class Phi(object):
    """Container class for dealing with ensemble residuals and phi vectors

    Parameters
    ----------
        em : EnsembleMethod
    """
    def __init__(self,em):
        assert isinstance(em,EnsembleMethod)
        self.em = em
        self.phi_files = {}
        self._tags = ["composite","meas","actual","reg"]
        self._prepare_output_files()
        self.obs0_matrix = self.em.obsensemble_0.nonzero.as_pyemu_matrix()
        if self.em.parensemble.istransformed:
            self.par0_matrix = self.em.parensemble_0.as_pyemu_matrix()
        else:
            self.par0_matrix = self.em.parensemble_0._transform(inplace=False).as_pyemu_matrix()
        self.par0_matrix = self.par0_matrix.get(col_names=self.em.parcov.row_names)
        self.em.logger.log("inverting parcov for regul phi calcs")
        self.inv_parcov = self.em.parcov.pseudo_inv(eigthresh=self.em.pst.svd_data.eigthresh)
        self.em.logger.log("inverting parcov for regul phi calcs")
        self.update()

    def _prepare_output_files(self):
        num_reals = self.em.obsensemble.shape[0]
        for tag in self._tags:
            f = open(self.em.pst.filename + ".iobj.{0}.csv".format(tag), 'w')
            f.write("iter_num,total_runs,lambda,min,max,mean,median,std,")
            f.write(','.join(["{0:010d}".format(i + 1) for i in range(num_reals)]))
            f.write('\n')
            self.phi_files[tag] = f

    def write(self,cur_lam=0.0):
        for t,pv in self.phi_vec_dict.items():
            self._phi_report(self.phi_files[t],pv,cur_lam)

    @property
    def phi_vec_dict(self):
        d = {t:pv for t,pv in zip(self._tags,[self.comp_phi,self.meas_phi,self.meas_phi_actual,self.reg_phi])}
        return d

    def _phi_report(self,phi_csv,phi_vec,cur_lam=0.0):
        phi_csv.write("{0},{1},{2},{3},{4},{5},{6},{7},".format(self.em.iter_num,
                                                             self.em.total_runs,
                                                             cur_lam,
                                                             phi_vec.min(),
                                                             phi_vec.max(),
                                                             phi_vec.mean(),
                                                             np.median(phi_vec),
                                                             phi_vec.std()))
        #[print(phi) for phi in phi_vec]
        phi_csv.write(",".join(["{0:20.8}".format(phi) for phi in phi_vec]))
        phi_csv.write("\n")
        phi_csv.flush()

    def get_meas_and_regul_phi(self,obsensemble,parensemble):
        assert obsensemble.shape[0] == parensemble.shape[0]
        meas_phi = self._calc_meas_phi(obsensemble)
        reg_phi = self._calc_regul_phi(parensemble)
        return meas_phi,reg_phi



    def update(self):
        #assert obsensemble.shape[0] == parensemble.shape[0]
        #self.cur_lam = cur_lam
        self.meas_phi = self._calc_meas_phi(self.em.obsensemble)
        self.meas_phi_actual = self._calc_meas_phi_actual(self.em.obsensemble)
        self.reg_phi = self._calc_regul_phi(self.em.parensemble)
        self.comp_phi = self.meas_phi + (self.reg_phi * self.em.regul_factor)


    # dynamic updating is too costly when lots of pars are being used...
    # @property
    # def meas_phi(self):
    #     #print('calc meas phi')
    #     return self._calc_meas_phi(self.em.obsensemble)
    #
    # @property
    # def meas_phi_actual(self):
    #     #print('calc actual phi')
    #     return self._calc_meas_phi_actual(self.em.obsensemble)
    #
    # @property
    # def reg_phi(self):
    #     #print('calc regul phi')
    #     return self._calc_regul_phi(self.em.parensemble)
    #
    # @property
    # def comp_phi(self):
    #     #print('calc comp phi')
    #     return self.meas_phi + (self.reg_phi * self.em.regul_factor)

    def report(self,cur_lam=0.0):
        self.write(cur_lam)
        ls = self.em.logger.statement
        pvd = self.phi_vec_dict
        ls("**** phi summary ****")
        ls("{0:12s}: {1:>15.6G}".format("regul factor",self.em.regul_factor))
        for t in self._tags:
            ls("{0:9s} mean: {1:>15.6G}, std: {2:>15.6G}".\
               format(t,pvd[t].mean(),pvd[t].std()))
        ls("*********************")

    def _calc_meas_phi(self,obsensemble):
        obs_diff = self.get_residual_obs_matrix(obsensemble)

        q = np.diagonal(self.em.obscov_inv_sqrt.get(row_names=obs_diff.col_names,col_names=obs_diff.col_names).x)
        phi_vec = []
        for i in range(obs_diff.shape[0]):
            o = obs_diff.x[i,:]
            phi_vec.append(((obs_diff.x[i,:] * q)**2).sum())

        return np.array(phi_vec)


    def _calc_regul_phi(self,parensemble):
        if isinstance(parensemble,pd.DataFrame):
            parensemble = pyemu.ParameterEnsemble.from_dataframe(pst=self.em.pst,df=parensemble)
        par_diff = self.get_residual_par_matrix(parensemble)
        #reg_phi_vec = np.diag(np.dot(np.dot(par_diff.x,self.inv_parcov.x),par_diff.x.transpose()))
        reg_phi_vec = np.diag((par_diff * self.inv_parcov * par_diff.T).x)
        return reg_phi_vec

    def _calc_meas_phi_actual(self,obsensemble):
        return obsensemble.phi_vector

    def get_residual_obs_matrix(self, obsensemble):
        obs_matrix = obsensemble.nonzero.as_pyemu_matrix()

        res_mat = obs_matrix - self.obs0_matrix.get(col_names=obs_matrix.col_names,
                                                    row_names=obs_matrix.row_names)
        self._apply_inequality_constraints(res_mat)
        return  res_mat

    def get_residual_par_matrix(self, parensemble):
        # these should already be transformed
        # if parensemble.istransformed:
        #     par_matrix = parensemble.as_pyemu_matrix()
        # else:
        #     par_matrix = parensemble._transform(inplace=False).as_pyemu_matrix()
        par_matrix = parensemble.as_pyemu_matrix()
        #res_mat = par_matrix.get(col_names=self.em.pst.adj_par_names) - \
        #          self.par0_matrix.get(col_names=self.em.pst.adj_par_names)
        res_mat = par_matrix.get(col_names=self.par0_matrix.col_names) - self.par0_matrix
        return  res_mat

    def _apply_inequality_constraints(self,res_mat):
        obs = self.em.pst.observation_data.loc[res_mat.col_names]
        gt_names = obs.loc[obs.obgnme.apply(lambda x: x.startswith("g_") or x.startswith("less")), "obsnme"]
        lt_names = obs.loc[obs.obgnme.apply(lambda x: x.startswith("l_") or x.startswith("greater")), "obsnme"]
        if gt_names.shape[0] == 0 and lt_names.shape[0] == 0:
            return res_mat
        res_df = res_mat.to_dataframe()
        if gt_names.shape[0] > 0:
            for gt_name in gt_names:
                #print(res_df.loc[:,gt_name])
                #if the residual is greater than zero, this means the ineq is satisified
                res_df.loc[res_df.loc[:,gt_name] > 0,gt_name] = 0.0
                #print(res_df.loc[:,gt_name])
                #print()


        if lt_names.shape[0] > 0:
            for lt_name in lt_names:
                #print(res_df.loc[:,lt_name])
                #f the residual is less than zero, this means the ineq is satisfied
                res_df.loc[res_df.loc[:,lt_name] < 0,lt_name] = 0.0
                #print(res_df.loc[:,lt_name])
                #print()





class EnsembleSmoother(EnsembleMethod):
    """an implementation of the GLM iterative ensemble smoother

    Parameters
    ----------
        pst : pyemu.Pst or str
            a control file instance or filename
        parcov : pyemu.Cov or str
            a prior parameter covariance matrix or filename. If None,
            parcov is constructed from parameter bounds (diagonal)
        obscov : pyemu.Cov or str
            a measurement noise covariance matrix or filename. If None,
            obscov is constructed from observation weights.
        num_slaves : int
            number of slaves to use in (local machine) parallel evaluation of the parmaeter
            ensemble.  If 0, serial evaluation is used.  Ignored if submit_file is not None
        use_approx_prior : bool
             a flag to use the MLE (approx) upgrade solution.  If True, a MAP
             solution upgrade is used
        submit_file : str
            the name of a HTCondor submit file.  If not None, HTCondor is used to
            evaluate the parameter ensemble in parallel by issuing condor_submit
            as a system command
        port : int
            the TCP port number to communicate on for parallel run management
        slave_dir : str
            path to a directory with a complete set of model files and PEST
            interface files
        drop_bad_reals : float
                drop realizations with phi greater than drop_bad_reals. If None, all
                realizations are kept. Default is None

    Example
    -------
    ``>>>import pyemu``

    ``>>>es = pyemu.EnsembleSmoother(pst="pest.pst")``
    """

    def __init__(self,pst,parcov=None,obscov=None,num_slaves=0,submit_file=None,verbose=False,
                 port=4004,slave_dir="template",drop_bad_reals=None,save_mats=False):



        super(EnsembleSmoother,self).__init__(pst=pst,parcov=parcov,obscov=obscov,num_slaves=num_slaves,
                                              submit_file=submit_file,verbose=verbose,port=port,
                                              slave_dir=slave_dir)
        self.logger.warn("pyemu's EnsembleSmoother is for prototyping only.  Use PESTPP-IES for a production " +
                      "implementation of iterative ensemble smoother")
        #self.use_approx_prior = bool(use_approx_prior)
        self.parcov_inv_sqrt = None
        self.half_obscov_diag = None
        self.delta_par_prior = None
        self.drop_bad_reals = drop_bad_reals
        self.save_mats = save_mats
    # @classmethod
    # def from_pestpp_args(cls,pst):
    #     if isinstance(pst,str):
    #         pst = pyemu.Pst(pst)
    #     pp = pst.pestpp_options
    #     parcov = pp.pop("parcov_filename",pyemu.Cov.from_parameter_data(pst,sigma_range=6))
    #     obscov = pyemu.Cov.from_observation_data(pst)
    #     par_csv = pp.pop("ies_par_csv")
    #     obs_csv = pp.pop("ies_obs_csv")
    #     restart_csv = pp.pop("ies_obs_restart_csv",None)
    #
    #     use_approx = False
    #     tags = ["ies_use_approx","ies_use_approximate_solution"]
    #     for tag in tags:
    #         if tag in pp.keys() and pp[tag].startswith('t'):
    #
    #             use_approx = True
    #             pp.pop(tag)
    #
    #     lambda_mults = pp.pop("lambda",[1.0])
    #     init_lambda = pp.pop("ies_init_lam",pp.pop("ies_initial_lambda",None))
    #     subset = pp.pop("ies_subset_size",None)




    def initialize(self,num_reals=1,init_lambda=None,enforce_bounds="reset",
                   parensemble=None,obsensemble=None,restart_obsensemble=None,
                   regul_factor=0.0,use_approx_prior=True,build_empirical_prior=False):
        """Initialize the iES process.  Depending on arguments, draws or loads
        initial parameter observations ensembles and runs the initial parameter
        ensemble

        Parameters
        ----------
            num_reals : int
                the number of realizations to draw.  Ignored if parensemble/obsensemble
                are not None
            init_lambda : float
                the initial lambda to use.  During subsequent updates, the lambda is
                updated according to upgrade success
            enforce_bounds : str
                how to enfore parameter bound transgression.  options are
                reset, drop, or None
            parensemble : pyemu.ParameterEnsemble or str
                a parameter ensemble or filename to use as the initial
                parameter ensemble.  If not None, then obsenemble must not be
                None
            obsensemble : pyemu.ObservationEnsemble or str
                an observation ensemble or filename to use as the initial
                observation ensemble.  If not None, then parensemble must
                not be None
            restart_obsensemble : pyemu.ObservationEnsemble or str
                an observation ensemble or filename to use as an
                evaluated observation ensemble.  If not None, this will skip the initial
                parameter ensemble evaluation - user beware!
            regul_factor : float
                the regularization penalty fraction of the composite objective.  The
                Prurist, MAP solution would be regul_factor = 1.0, yielding equal
                parts measurement and regularization to the composite objective function.
                Default is 0.0, which means only seek to minimize the measurement objective
                function
            use_approx_prior : bool
                a flag to use the inverse, square root of the prior ccovariance matrix
                for scaling the upgrade calculation.  If True, this matrix is not used.
                Default is True
            build_empirical_prior : bool
                flag to build the prior parameter covariance matrix from an existing parensemble.
                If True and parensemble is None, an exception is raised


        Example
        -------
        ``>>>import pyemu``

        ``>>>es = pyemu.EnsembleSmoother(pst="pest.pst")``

        ``>>>es.initialize(num_reals=100)``

        """
        '''
        (re)initialize the process
        '''
        # initialize the phi report csv
        self.enforce_bounds = enforce_bounds

        self.regul_factor = float(regul_factor)

        self.total_runs = 0
        # this matrix gets used a lot, so only calc once and store
        self.obscov_inv_sqrt = self.obscov.get(self.pst.nnz_obs_names).inv.sqrt

        if use_approx_prior:
            self.logger.statement("using approximate parcov in solution")
            self.parcov_inv_sqrt = 1.0
        else:
            self.logger.statement("using full parcov in solution")
            # Chen and Oliver use a low rank approx here, but so far,
            # I haven't needed it - not using enough parameters yet
            self.logger.log("forming inverse sqrt parcov matrix")
            self.parcov_inv_sqrt = self.parcov.inv.sqrt
            self.logger.log("forming inverse sqrt parcov matrix")

        if parensemble is not None and obsensemble is not None:
            self.logger.log("initializing with existing ensembles")
            if isinstance(parensemble,str):
                self.logger.log("loading parensemble from file")
                if not os.path.exists(obsensemble):
                    self.logger.lraise("can not find parensemble file: {0}".\
                                       format(parensemble))
                df = pd.read_csv(parensemble,index_col=0)
                #df.index = [str(i) for i in df.index]
                self.parensemble_0 = ParameterEnsemble.from_dataframe(df=df,pst=self.pst)
                self.logger.log("loading parensemble from file")

            elif isinstance(parensemble,ParameterEnsemble):
                self.parensemble_0 = parensemble.copy()
            else:
                raise Exception("unrecognized arg type for parensemble, " +\
                                "should be filename or ParameterEnsemble" +\
                                ", not {0}".format(type(parensemble)))
            self.parensemble = self.parensemble_0.copy()
            if isinstance(obsensemble,str):
                self.logger.log("loading obsensemble from file")
                if not os.path.exists(obsensemble):
                    self.logger.lraise("can not find obsensemble file: {0}".\
                                       format(obsensemble))
                df = pd.read_csv(obsensemble,index_col=0).loc[:,self.pst.nnz_obs_names]
                #df.index = [str(i) for i in df.index]
                self.obsensemble_0 = ObservationEnsemble.from_dataframe(df=df,pst=self.pst)
                self.logger.log("loading obsensemble from file")

            elif isinstance(obsensemble,ObservationEnsemble):
                self.obsensemble_0 = obsensemble.copy()
            else:
                raise Exception("unrecognized arg type for obsensemble, " +\
                                "should be filename or ObservationEnsemble" +\
                                ", not {0}".format(type(obsensemble)))

            assert self.parensemble_0.shape[0] == self.obsensemble_0.shape[0]
            #self.num_reals = self.parensemble_0.shape[0]
            num_reals = self.parensemble.shape[0]
            self.logger.log("initializing with existing ensembles")

            if build_empirical_prior:

                self.reset_parcov(self.parensemble.covariance_matrix())
                if self.save_mats:
                    self.parcov.to_binary(self.pst.filename+".empcov.jcb")

        else:
            if build_empirical_prior:
                self.logger.lraise("can't use build_emprirical_prior without parensemble...")
            self.logger.log("initializing smoother with {0} realizations".format(num_reals))
            self.logger.log("initializing parensemble")
            self.parensemble_0 = pyemu.ParameterEnsemble.from_gaussian_draw(self.pst,
                                                                            self.parcov,num_reals=num_reals)
            self.parensemble_0.enforce(enforce_bounds=enforce_bounds)
            self.logger.log("initializing parensemble")
            self.parensemble = self.parensemble_0.copy()
            self.parensemble_0.to_csv(self.pst.filename +\
                                      self.paren_prefix.format(0))
            self.logger.log("initializing parensemble")
            self.logger.log("initializing obsensemble")
            self.obsensemble_0 = pyemu.ObservationEnsemble.from_id_gaussian_draw(self.pst,
                                                                                 num_reals=num_reals)
            #self.obsensemble = self.obsensemble_0.copy()

            # save the base obsensemble
            self.obsensemble_0.to_csv(self.pst.filename +\
                                      self.obsen_prefix.format(-1))
            self.logger.log("initializing obsensemble")
            self.logger.log("initializing smoother with {0} realizations".format(num_reals))


        if use_approx_prior:
            self.logger.statement("using approximate parcov in solution")
            self.parcov_inv_sqrt = 1.0
        else:
            self.logger.statement("using full parcov in solution")
            # Chen and Oliver use a low rank approx here, but so far,
            # I haven't needed it - not using enough parameters yet
            self.logger.log("forming inverse sqrt parcov matrix")
            self.parcov_inv_sqrt = self.parcov.inv.sqrt
            self.logger.log("forming inverse sqrt parcov matrix")

        # self.obs0_matrix = self.obsensemble_0.nonzero.as_pyemu_matrix()
        # self.par0_matrix = self.parensemble_0.as_pyemu_matrix()
        self.enforce_bounds = enforce_bounds

        if restart_obsensemble is not None:
            self.logger.log("loading restart_obsensemble {0}".format(restart_obsensemble))
            failed_runs,self.obsensemble = self._load_obs_ensemble(restart_obsensemble)
            assert self.obsensemble.shape[0] == self.obsensemble_0.shape[0]
            assert list(self.obsensemble.columns) == list(self.obsensemble_0.columns)
            self.logger.log("loading restart_obsensemble {0}".format(restart_obsensemble))

        else:
            # run the initial parameter ensemble
            self.logger.log("evaluating initial ensembles")
            failed_runs, self.obsensemble = self._calc_obs(self.parensemble)
            self.obsensemble.to_csv(self.pst.filename +\
                                      self.obsen_prefix.format(0))
            if self.raw_sweep_out is not None:
                self.raw_sweep_out.to_csv(self.pst.filename + "_sweepraw0.csv")
            self.logger.log("evaluating initial ensembles")

        if failed_runs is not None:
            self.logger.warn("dropping failed realizations")
            #failed_runs_str = [str(f) for f in failed_runs]
            #self.parensemble = self.parensemble.drop(failed_runs)
            #self.obsensemble = self.obsensemble.drop(failed_runs)
            self.parensemble.loc[failed_runs,:] = np.NaN
            self.parensemble = self.parensemble.dropna()
            self.obsensemble.loc[failed_runs,:] = np.NaN
            self.obsensemble = self.obsensemble.dropna()

        if not self.parensemble.istransformed:
            self.parensemble._transform(inplace=True)
        if not self.parensemble_0.istransformed:
            self.parensemble_0._transform(inplace=True)

        self.phi = Phi(self)

        if self.drop_bad_reals is not None:
            #drop_idx = np.argwhere(self.current_phi_vec > self.drop_bad_reals).flatten()
            #comp_phi = self.phi.comp_phi
            #drop_idx = np.argwhere(self.phi.comp_phi > self.drop_bad_reals).flatten()
            #meas_phi = self.phi.meas_phi
            drop_idx = np.argwhere(self.phi.meas_phi > self.drop_bad_reals).flatten()
            run_ids = self.obsensemble.index.values
            drop_idx = run_ids[drop_idx]
            if len(drop_idx) == self.obsensemble.shape[0]:
                raise Exception("dropped all realizations as 'bad'")
            if len(drop_idx) > 0:
                self.logger.warn("{0} realizations dropped as 'bad' (indices :{1})".\
                                 format(len(drop_idx),','.join([str(d) for d in drop_idx])))
                self.parensemble.loc[drop_idx,:] = np.NaN
                self.parensemble = self.parensemble.dropna()
                self.obsensemble.loc[drop_idx,:] = np.NaN
                self.obsensemble = self.obsensemble.dropna()

                self.phi.update()

        self.phi.report(cur_lam=0.0)

        self.last_best_mean = self.phi.comp_phi.mean()
        self.last_best_std = self.phi.comp_phi.std()

        #self.logger.statement("initial phi (mean, std): {0:15.6G},{1:15.6G}".\
        #                      format(self.last_best_mean,self.last_best_std))
        if init_lambda is not None:
            self.current_lambda = float(init_lambda)
        else:
            #following chen and oliver
            x = self.last_best_mean / (2.0 * float(self.obsensemble.shape[1]))
            self.current_lambda = 10.0**(np.floor(np.log10(x)))

        self.logger.statement("current lambda:{0:15.6g}".format(self.current_lambda))

        self.delta_par_prior = self._calc_delta_par(self.parensemble_0)
        u,s,v = self.delta_par_prior.pseudo_inv_components(eigthresh=self.pst.svd_data.eigthresh)
        self.Am = u * s.inv
        if self.save_mats:
            np.savetxt(self.pst.filename.replace(".pst",'.') + "0.prior_par_diff.dat", self.delta_par_prior.x, fmt="%15.6e")
            np.savetxt(self.pst.filename.replace(".pst",'.') + "0.am_u.dat",u.x,fmt="%15.6e")
            np.savetxt(self.pst.filename.replace(".pst", '.') + "0.am_v.dat", v.x, fmt="%15.6e")
            np.savetxt(self.pst.filename.replace(".pst",'.') + "0.am_s_inv.dat", s.inv.as_2d, fmt="%15.6e")
            np.savetxt(self.pst.filename.replace(".pst",'.') + "0.am.dat", self.Am.x, fmt="%15.6e")

        self._initialized = True

    def get_localizer(self):
        """ get an empty/generic localizer matrix that can be filled

        Returns
        -------
            localizer : pyemu.Matrix
                matrix with nnz obs names for rows and adj par names for columns

        """
        onames = self.pst.nnz_obs_names
        pnames = self.pst.adj_par_names
        localizer = Matrix(x=np.ones((len(onames),len(pnames))),row_names=onames,col_names=pnames)
        return localizer

    def _calc_delta_par(self,parensemble):
        '''
        calc the scaled parameter ensemble differences from the mean
        '''
        return self._calc_delta(parensemble, self.parcov_inv_sqrt)

    def _calc_delta_obs(self,obsensemble):
        '''
        calc the scaled observation ensemble differences from the mean
        '''
        return self._calc_delta(obsensemble.nonzero, self.obscov_inv_sqrt)


    def update(self,lambda_mults=[1.0],localizer=None,run_subset=None,use_approx=True,
               calc_only=False):
        """update the iES one GLM cycle

        Parameters
        ----------
            lambda_mults : list
                a list of lambda multipliers to test.  Each lambda mult value will require
                evaluating (a subset of) the parameter ensemble.
            localizer : pyemu.Matrix
                a jacobian localizing matrix
            run_subset : int
                the number of realizations to test for each lambda_mult value.  For example,
                if run_subset = 30 and num_reals=100, the first 30 realizations will be run (in
                parallel) for each lambda_mult value.  Then the best lambda_mult is selected and the
                remaining 70 realizations for that lambda_mult value are run (in parallel).
            use_approx : bool
                 a flag to use the MLE or MAP upgrade solution.  True indicates use MLE solution
            calc_only : bool
                a flag to calculate the upgrade matrix only (not run the ensemble). This is mostly for
                debugging and testing on travis. Default is False

        Example
        -------

        ``>>>import pyemu``

        ``>>>es = pyemu.EnsembleSmoother(pst="pest.pst")``

        ``>>>es.initialize(num_reals=100)``

        ``>>>es.update(lambda_mults=[0.1,1.0,10.0],run_subset=30)``

         """

        #if not self.parensemble.istransformed:
        #    self.parensemble._transform(inplace=False)

        if run_subset is not None:
            if run_subset >= self.obsensemble.shape[0]:
                self.logger.warn("run_subset ({0}) >= num of active reals ({1})...ignoring ".\
                                 format(run_subset,self.obsensemble.shape[0]))
                run_subset = None

        self.iter_num += 1
        mat_prefix =  self.pst.filename.replace('.pst','')+".{0}".format(self.iter_num)
        self.logger.log("iteration {0}".format(self.iter_num))
        self.logger.statement("{0} active realizations".format(self.obsensemble.shape[0]))
        if self.obsensemble.shape[0] < 2:
            self.logger.lraise("at least active 2 realizations (really like 300) are needed to update")
        if not self._initialized:
            #raise Exception("must call initialize() before update()")
            self.logger.lraise("must call initialize() before update()")

        self.logger.log("calculate scaled delta obs")
        scaled_delta_obs = self._calc_delta_obs(self.obsensemble)
        self.logger.log("calculate scaled delta obs")
        self.logger.log("calculate scaled delta par")
        scaled_delta_par = self._calc_delta_par(self.parensemble)
        self.logger.log("calculate scaled delta par")

        self.logger.log("calculate pseudo inv comps")
        u,s,v = scaled_delta_obs.pseudo_inv_components(eigthresh=self.pst.svd_data.eigthresh)
        s.col_names = s.row_names
        self.logger.log("calculate pseudo inv comps")

        self.logger.log("calculate obs diff matrix")
        #obs_diff = self.obscov_inv_sqrt * self._get_residual_obs_matrix(self.obsensemble).T
        obs_diff = self.obscov_inv_sqrt * self.phi.get_residual_obs_matrix(self.obsensemble).T
        self.logger.log("calculate obs diff matrix")

        if self.save_mats:
            np.savetxt(mat_prefix + ".obs_diff.dat", scaled_delta_obs.x, fmt="%15.6e")
            np.savetxt(mat_prefix + ".par_diff.dat", scaled_delta_par.x, fmt="%15.6e")
            np.savetxt(mat_prefix + ".u.dat", u.x, fmt="%15.6e")
            np.savetxt(mat_prefix + ".s.dat", s.x, fmt="%15.6e")
            np.savetxt(mat_prefix + ".v.dat", v.x, fmt="%15.6e")
        # here is the math part...calculate upgrade matrices
        mean_lam,std_lam,paren_lam,obsen_lam = [],[],[],[]
        lam_vals = []
        for ilam,cur_lam_mult in enumerate(lambda_mults):

            parensemble_cur_lam = self.parensemble.copy()
            #print(parensemble_cur_lam.isnull().values.any())

            cur_lam = self.current_lambda * cur_lam_mult
            lam_vals.append(cur_lam)
            self.logger.log("calcs for  lambda {0}".format(cur_lam_mult))
            scaled_ident = Cov.identity_like(s) * (cur_lam+1.0)
            scaled_ident += s**2
            scaled_ident = scaled_ident.inv

            # build up this matrix as a single element so we can apply
            # localization
            self.logger.log("building upgrade_1 matrix")
            upgrade_1 = -1.0 * (self.parcov_inv_sqrt * scaled_delta_par) *\
                        v * s * scaled_ident * u.T
            if self.save_mats:
                np.savetxt(mat_prefix+".ivec.dat".format(self.iter_num), scaled_ident.x, fmt="%15.6e")
            self.logger.log("building upgrade_1 matrix")

            # apply localization
            if localizer is not None:
                self.logger.log("applying localization")
                upgrade_1.hadamard_product(localizer)
                self.logger.log("applying localization")

            # apply residual information
            self.logger.log("applying residuals")
            upgrade_1 *= obs_diff
            self.logger.log("applying residuals")

            self.logger.log("processing upgrade_1")
            if self.save_mats:
                np.savetxt(mat_prefix + ".upgrade_1.dat", upgrade_1.T.x, fmt="%15.6e")
            upgrade_1 = upgrade_1.to_dataframe()
            upgrade_1.index.name = "parnme"
            upgrade_1 = upgrade_1.T
            upgrade_1.index = [int(i) for i in upgrade_1.index]
            upgrade_1.to_csv(self.pst.filename+".upgrade_1.{0:04d}.csv".\
                               format(self.iter_num))
            if upgrade_1.isnull().values.any():
                    self.logger.lraise("NaNs in upgrade_1")
            self.logger.log("processing upgrade_1")

            #print(upgrade_1.isnull().values.any())
            #print(parensemble_cur_lam.index)
            #print(upgrade_1.index)
            parensemble_cur_lam += upgrade_1

            # parameter-based upgrade portion
            if not use_approx and self.iter_num > 1:
            #if True:
                self.logger.log("building upgrade_2 matrix")
                par_diff = (self.parensemble - self.parensemble_0.loc[self.parensemble.index,:]).\
                    as_pyemu_matrix().T
                x4 = self.Am.T * self.parcov_inv_sqrt * par_diff
                x5 = self.Am * x4
                x6 = scaled_delta_par.T * x5
                x7 = v * scaled_ident * v.T * x6
                ug2_mat = -1.0 * (self.parcov_inv_sqrt *
                                   scaled_delta_par * x7)
                upgrade_2 = ug2_mat.to_dataframe()
                upgrade_2.index.name = "parnme"
                upgrade_2 = upgrade_2.T
                upgrade_2.to_csv(self.pst.filename+".upgrade_2.{0:04d}.csv".\
                                   format(self.iter_num))
                upgrade_2.index = [int(i) for i in upgrade_2.index]

                if self.save_mats:
                    np.savetxt(mat_prefix + ".scaled_par_resid.dat", par_diff.x, fmt="%15.6e")
                    np.savetxt(mat_prefix + ".x4.dat", x4.x, fmt="%15.6e")
                    np.savetxt(mat_prefix + ".x5.dat", x5.x, fmt="%15.6e")
                    np.savetxt(mat_prefix + ".x6.dat", x6.x, fmt="%15.6e")
                    np.savetxt(mat_prefix + ".x7.dat", x7.x, fmt="%15.6e")
                    np.savetxt(mat_prefix + ".upgrade_2.dat", ug2_mat.T.x, fmt="%15.6e")

                if upgrade_2.isnull().values.any():
                    self.logger.lraise("NaNs in upgrade_2")

                parensemble_cur_lam += upgrade_2
                self.logger.log("building upgrade_2 matrix")
            self.logger.log("enforcing bounds")
            parensemble_cur_lam.enforce(self.enforce_bounds)
            self.logger.log("enforcing bounds")

            self.logger.log("filling fixed parameters")
            #fill in fixed pars with initial values
            fi = parensemble_cur_lam.fixed_indexer
            li = parensemble_cur_lam.log_indexer
            log_values = self.pst.parameter_data.loc[:,"parval1"].copy()
            log_values.loc[li] = log_values.loc[li].apply(np.log10)
            fixed_vals = log_values.loc[fi]

            for fname, fval in zip(fixed_vals.index, fixed_vals.values):
                # if fname not in df.columns:
                #    continue
                # print(fname)
                parensemble_cur_lam.loc[:, fname] = fval
            self.logger.log("filling fixed parameters")
            # this is for testing failed runs on upgrade testing
            # works with the 10par_xsec smoother test
            #parensemble_cur_lam.iloc[:,:] = -1000000.0

            # some hackery - we lose track of the transform flag here, but just
            # know it is transformed.  Need to create dataframe here because
            # pd.concat doesn't like par ensembles later
            paren_lam.append(pd.DataFrame(parensemble_cur_lam.loc[:,:]))
            self.logger.log("calcs for  lambda {0}".format(cur_lam_mult))

        if calc_only:
            return


        # subset if needed
        # and combine lambda par ensembles into one par ensemble for evaluation
        if run_subset is not None and run_subset < self.parensemble.shape[0]:
            #subset_idx = ["{0:d}".format(i) for i in np.random.randint(0,self.parensemble.shape[0]-1,run_subset)]
            subset_idx = self.parensemble.iloc[:run_subset,:].index.values
            self.logger.statement("subset idxs: " + ','.join([str(s) for s in subset_idx]))

            # more tracking of transformed - just know it! Creating dataframes...
            paren_lam_subset = [pe.loc[subset_idx,:] for pe in paren_lam]
            paren_combine = pd.concat(paren_lam_subset,ignore_index=True)
            paren_lam_subset = None
        else:
            subset_idx = self.parensemble.index.values
            paren_combine = pd.concat(paren_lam,ignore_index=True)


        self.logger.log("evaluating ensembles for lambdas : {0}".\
                        format(','.join(["{0:8.3E}".format(l) for l in lam_vals])))
        # back to par ensemble and know it is transformed
        paren_combine = ParameterEnsemble.from_dataframe(df=paren_combine,pst=self.pst,istransformed=True)
        failed_runs, obsen_combine = self._calc_obs(paren_combine)
        self.logger.log("evaluating ensembles for lambdas : {0}".\
                        format(','.join(["{0:8.3E}".format(l) for l in lam_vals])))
        paren_combine = None

        if failed_runs is not None and len(failed_runs) == obsen_combine.shape[0]:
                self.logger.lraise("all runs failed - cannot continue")


        # unpack lambda obs ensembles from combined obs ensemble
        nrun_per_lam = self.obsensemble.shape[0]
        if run_subset is not None:
            nrun_per_lam = run_subset
        obsen_lam = []

        for i in range(len(lam_vals)):
            sidx = i * nrun_per_lam
            eidx = sidx + nrun_per_lam
            oe = ObservationEnsemble.from_dataframe(df=obsen_combine.iloc[sidx:eidx,:].copy(),
                                                    pst=self.pst)
            oe.index = subset_idx
            # check for failed runs in this set - drop failed runs from obs ensembles
            if failed_runs is not None:
                failed_runs_this = np.array([f for f in failed_runs if f >= sidx and f < eidx]) - sidx
                if len(failed_runs_this) > 0:
                    if len(failed_runs_this) == oe.shape[0]:
                        self.logger.warn("all runs failed for lambda {0}".format(lam_vals[i]))
                    else:
                        self.logger.warn("{0} run failed for lambda {1}".\
                                         format(len(failed_runs_this),lam_vals[i]))
                    oe.iloc[failed_runs_this,:] = np.NaN
                    oe = oe.dropna()
                    paren_lam[i].iloc[failed_runs_this,:] = np.NaN
                    paren_lam[i] = ParameterEnsemble.from_dataframe(df=paren_lam[i].dropna(),pst=self.pst)
                    paren_lam[i].__instransformed = True

            # don't drop bad reals here, instead, mask bad reals in the lambda
            # selection and drop later
            # if self.drop_bad_reals is not None:
            #     assert isinstance(drop_bad_reals, float)
            #     drop_idx = np.argwhere(self.current_phi_vec > self.drop_bad_reals).flatten()
            #     run_ids = self.obsensemble.index.values
            #     drop_idx = run_ids[drop_idx]
            #     if len(drop_idx) == self.obsensemble.shape[0]:
            #         raise Exception("dropped all realizations as 'bad'")
            #     if len(drop_idx) > 0:
            #         self.logger.warn("{0} realizations dropped as 'bad' (indices :{1})". \
            #                          format(len(drop_idx), ','.join([str(d) for d in drop_idx])))
            #         self.parensemble.loc[drop_idx, :] = np.NaN
            #         self.parensemble = self.parensemble.dropna()
            #         self.obsensemble.loc[drop_idx, :] = np.NaN
            #         self.obsensemble = self.obsensemble.dropna()
            #
            #         self.current_phi_vec = self._calc_phi_vec(self.obsensemble)

            obsen_lam.append(oe)
        obsen_combine = None

        # here is where we need to select out the "best" lambda par and obs
        # ensembles

        #phi_vecs = [self._calc_phi_vec(obsen) for obsen in obsen_lam]
        #phi_vecs_reg = [self._calc_regul_phi_vec(paren) for paren in paren_lam]
        #if self.regul_factor > 0.0:
        #    for i,(pv,prv) in enumerate(zip(phi_vecs,phi_vecs_reg)):
        #        phi_vecs[i] = pv + (prv * self.regul_factor)
        self.logger.log("calc lambda phi vectors")
        phi_vecs = [self.phi.get_meas_and_regul_phi(oe,pe.loc[oe.index,:]) for oe,pe in zip(obsen_lam,paren_lam)]
        self.logger.log("calc lambda phi vectors")
        if self.drop_bad_reals is not None:
            for i,(meas_pv,regul_pv) in enumerate(phi_vecs):
                #for testing the drop_bad_reals functionality
                #pv[[0,3,7]] = self.drop_bad_reals + 1.0
                regul_pv = regul_pv.copy()
                regul_pv[meas_pv>self.drop_bad_reals] = np.NaN
                regul_pv = regul_pv[~np.isnan(regul_pv)]
                meas_pv[meas_pv>self.drop_bad_reals] = np.NaN
                meas_pv = meas_pv[~np.isnan(meas_pv)]
                if len(meas_pv) == 0:
                    #raise Exception("all realization for lambda {0} dropped as 'bad'".\
                    #                format(lam_vals[i]))
                    self.logger.warn("all realizations for lambda {0} marked as 'bad'")
                    meas_pv = np.zeros_like(obsen_lam[0].shape[0]) + 1.0e+30
                    regul_pv = np.zeros_like(obsen_lam[0].shape[0]) + 1.0e+30
                phi_vecs[i] = (meas_pv,regul_pv)
        mean_std_meas = [(pv[0].mean(),pv[0].std()) for pv in phi_vecs]
        mean_std_regul = [(pv[1].mean(), pv[1].std()) for pv in phi_vecs]
        update_pars = False
        update_lambda = False
        self.logger.statement("**************************")
        # self.logger.statement(str(datetime.now()))
        self.logger.statement("lambda testing summary")
        self.logger.statement("total runs:{0}".format(self.total_runs))
        self.logger.statement("iteration: {0}".format(self.iter_num))
        self.logger.statement("current lambda:{0:15.6G}, mean:{1:15.6G}, std:{2:15.6G}". \
                              format(self.current_lambda,
                                     self.last_best_mean, self.last_best_std))

        # accept a new best if its within 10%
        best_mean = self.last_best_mean * 1.1
        best_std = self.last_best_std * 1.1
        best_i = 0
        for i,((mm,ms),(rm,rs)) in enumerate(zip(mean_std_meas,mean_std_regul)):
            self.logger.statement(" tested lambda:{0:15.6G}, meas mean:{1:15.6G}, meas std:{2:15.6G}".
                                  format(self.current_lambda * lambda_mults[i],mm,ms))
            self.logger.statement("{0:30s}regul mean:{1:15.6G}, regul std:{2:15.6G}".\
                                  format(' ',rm,rs))
            m = mm + (self.regul_factor * rm)
            s = ms + (self.regul_factor * rs)
            if m < best_mean:
                update_pars = True
                best_mean = m
                best_i = i
                if s < best_std:
                    update_lambda = True
                    best_std = s
        if np.isnan(best_mean):
            self.logger.lraise("best mean = NaN")
        if np.isnan(best_std):
            self.logger.lraise("best std = NaN")

        if not update_pars:
            self.current_lambda *= max(lambda_mults) * 10.0
            self.current_lambda = min(self.current_lambda,100000)
            self.logger.statement("not accepting iteration, increased lambda:{0}".\
                  format(self.current_lambda))
        else:
            #more transformation status hard coding - ugly
            self.parensemble = ParameterEnsemble.from_dataframe(df=paren_lam[best_i],pst=self.pst,istransformed=True)
            if run_subset is not None:
                failed_runs, self.obsensemble = self._calc_obs(self.parensemble)
                if failed_runs is not None:
                    self.logger.warn("dropping failed realizations")
                    self.parensemble.loc[failed_runs, :] = np.NaN
                    self.parensemble = self.parensemble.dropna()
                    self.obsensemble.loc[failed_runs, :] = np.NaN
                    self.obsensemble = self.obsensemble.dropna()

                self.phi.update()
                best_mean = self.phi.comp_phi.mean()
                best_std = self.phi.comp_phi.std()
            else:
                self.obsensemble = obsen_lam[best_i]
                # reindex parensemble in case failed runs
                self.parensemble = ParameterEnsemble.from_dataframe(df=self.parensemble.loc[self.obsensemble.index],
                                                                    pst=self.pst,
                                                                    istransformed=self.parensemble.istransformed)
                self.phi.update()
            if self.drop_bad_reals is not None:
                # for testing drop_bad_reals functionality
                # self.current_phi_vec[::2] = self.drop_bad_reals + 1.0
                #drop_idx = np.argwhere(self.current_phi_vec > self.drop_bad_reals).flatten()
                drop_idx = np.argwhere(self.phi.comp_phi > self.drop_bad_reals).flatten()
                run_ids = self.obsensemble.index.values
                drop_idx = run_ids[drop_idx]
                if len(drop_idx) > self.obsensemble.shape[0] - 3:
                    raise Exception("dropped too many realizations as 'bad'")
                if len(drop_idx) > 0:
                    self.logger.warn("{0} realizations dropped as 'bad' (indices :{1})". \
                                     format(len(drop_idx), ','.join([str(d) for d in drop_idx])))
                    self.parensemble.loc[drop_idx, :] = np.NaN
                    self.parensemble = self.parensemble.dropna()
                    self.obsensemble.loc[drop_idx, :] = np.NaN
                    self.obsensemble = self.obsensemble.dropna()

                    self.phi.update()
                    best_mean = self.phi.comp_phi.mean()
                    best_std = self.phi.comp_phi.std()

            self.phi.report(cur_lam=self.current_lambda * lambda_mults[best_i])

            self.logger.statement("   best lambda:{0:15.6G}, mean:{1:15.6G}, std:{2:15.6G}".\
                  format(self.current_lambda*lambda_mults[best_i],
                         best_mean,best_std))
            #self.logger.statement("   actual mean phi: {0:15.6G}".format(float(self.current_actual_phi.mean())))
            self.last_best_mean = best_mean
            self.last_best_std = best_std


        if update_lambda:
            # be aggressive
            self.current_lambda *= (lambda_mults[best_i] * 0.75)
            # but don't let lambda get too small
            self.current_lambda = max(self.current_lambda,0.00001)
            self.logger.statement("updating lambda: {0:15.6G}".\
                  format(self.current_lambda ))

        self.logger.statement("**************************\n")
        self.parensemble.to_csv(self.pst.filename+self.paren_prefix.\
                                    format(self.iter_num))
        self.obsensemble.to_csv(self.pst.filename+self.obsen_prefix.\
                                    format(self.iter_num))
        if self.raw_sweep_out is not None:
            self.raw_sweep_out.to_csv(self.pst.filename+"_sweepraw{0}.csv".\
                                        format(self.iter_num))
        self.logger.log("iteration {0}".format(self.iter_num))



"""A base class for developing prototype ensemble methods
"""
from __future__ import print_function, division
import os
from datetime import datetime
import shutil
import threading
import time
import warnings

import numpy as np
import pandas as pd
import pyemu
from pyemu.en import ParameterEnsemble,ObservationEnsemble
from pyemu.mat import Cov,Matrix

from pyemu.pst import Pst
from ..logger import Logger



class EnsembleMethod(object):
    """Base class for ensemble-type methods.  Should not be instantiated directly

    Parameters
    ----------
        pst : pyemu.Pst or str
            a control file instance or filename
        parcov : pyemu.Cov or str
            a prior parameter covariance matrix or filename. If None,
            parcov is constructed from parameter bounds (diagonal)
        obscov : pyemu.Cov or str
            a measurement noise covariance matrix or filename. If None,
            obscov is constructed from observation weights.
        num_slaves : int
            number of slaves to use in (local machine) parallel evaluation of the parmaeter
            ensemble.  If 0, serial evaluation is used.  Ignored if submit_file is not None
        submit_file : str
            the name of a HTCondor submit file.  If not None, HTCondor is used to
            evaluate the parameter ensemble in parallel by issuing condor_submit
            as a system command
        port : int
            the TCP port number to communicate on for parallel run management
        slave_dir : str
            path to a directory with a complete set of model files and PEST
            interface files

    """

    def __init__(self,pst,parcov=None,obscov=None,num_slaves=0,use_approx_prior=True,
                 submit_file=None,verbose=False,port=4004,slave_dir="template"):
        self.logger = Logger(verbose)
        if verbose is not False:
            self.logger.echo = True
        self.num_slaves = int(num_slaves)
        if submit_file is not None:
            if not os.path.exists(submit_file):
                self.logger.lraise("submit_file {0} not found".format(submit_file))
        elif num_slaves > 0:
            if not os.path.exists(slave_dir):
                self.logger.lraise("template dir {0} not found".format(slave_dir))

        self.slave_dir = slave_dir
        self.submit_file = submit_file
        self.port = int(port)
        self.paren_prefix = ".parensemble.{0:04d}.csv"
        self.obsen_prefix = ".obsensemble.{0:04d}.csv"

        if isinstance(pst,str):
            pst = Pst(pst)
        assert isinstance(pst,Pst)
        self.pst = pst
        self.sweep_in_csv = pst.pestpp_options.get("sweep_parameter_csv_file","sweep_in.csv")
        self.sweep_out_csv = pst.pestpp_options.get("sweep_output_csv_file","sweep_out.csv")
        if parcov is not None:
            assert isinstance(parcov,Cov)
        else:
            parcov = Cov.from_parameter_data(self.pst)
        if obscov is not None:
            assert isinstance(obscov,Cov)
        else:
            obscov = Cov.from_observation_data(pst)

        self.parcov = parcov
        self.obscov = obscov

        self._initialized = False
        self.iter_num = 0
        self.total_runs = 0
        self.raw_sweep_out = None

    def initialize(self,*args,**kwargs):
        raise Exception("EnsembleMethod.initialize() must be implemented by the derived types")

    def _calc_delta(self,ensemble,scaling_matrix=None):
        '''
        calc the scaled  ensemble differences from the mean
        '''
        mean = np.array(ensemble.mean(axis=0))
        delta = ensemble.as_pyemu_matrix()
        for i in range(ensemble.shape[0]):
            delta.x[i,:] -= mean
        if scaling_matrix is not None:
            delta = scaling_matrix * delta.T
        delta *= (1.0 / np.sqrt(float(ensemble.shape[0] - 1.0)))
        return delta

    def _calc_obs(self,parensemble):
        self.logger.log("removing existing sweep in/out files")
        try:
            os.remove(self.sweep_in_csv)
        except Exception as e:
            self.logger.warn("error removing existing sweep in file:{0}".format(str(e)))
        try:
            os.remove(self.sweep_out_csv)
        except Exception as e:
            self.logger.warn("error removing existing sweep out file:{0}".format(str(e)))
        self.logger.log("removing existing sweep in/out files")

        if parensemble.isnull().values.any():
            parensemble.to_csv("_nan.csv")
            self.logger.lraise("_calc_obs() error: NaNs in parensemble (written to '_nan.csv')")

        if self.submit_file is None:
            self._calc_obs_local(parensemble)
        else:
            self._calc_obs_condor(parensemble)

        # make a copy of sweep out for restart purposes
        # sweep_out = str(self.iter_num)+"_raw_"+self.sweep_out_csv
        # if os.path.exists(sweep_out):
        #     os.remove(sweep_out)
        # shutil.copy2(self.sweep_out_csv,sweep_out)

        self.logger.log("reading sweep out csv {0}".format(self.sweep_out_csv))
        failed_runs,obs = self._load_obs_ensemble(self.sweep_out_csv)
        self.logger.log("reading sweep out csv {0}".format(self.sweep_out_csv))
        self.total_runs += obs.shape[0]
        self.logger.statement("total runs:{0}".format(self.total_runs))
        return failed_runs,obs

    def _load_obs_ensemble(self,filename):
        if not os.path.exists(filename):
            self.logger.lraise("obsensemble file {0} does not exists".format(filename))
        obs = pd.read_csv(filename)
        obs.columns = [item.lower() for item in obs.columns]
        self.raw_sweep_out = obs.copy() # save this for later to support restart
        assert "input_run_id" in obs.columns,\
            "'input_run_id' col missing...need newer version of sweep"
        obs.index = obs.input_run_id
        failed_runs = None
        if 1 in obs.failed_flag.values:
            failed_runs = obs.loc[obs.failed_flag == 1].index.values
            self.logger.warn("{0} runs failed (indices: {1})".\
                             format(len(failed_runs),','.join([str(f) for f in failed_runs])))
        obs = ObservationEnsemble.from_dataframe(df=obs.loc[:,self.obscov.row_names],
                                                               pst=self.pst)
        if obs.isnull().values.any():
            self.logger.lraise("_calc_obs() error: NaNs in obsensemble")
        return failed_runs, obs

    def _get_master_thread(self):
        master_stdout = "_master_stdout.dat"
        master_stderr = "_master_stderr.dat"
        def master():
            try:
                #os.system("sweep {0} /h :{1} 1>{2} 2>{3}". \
                #          format(self.pst.filename, self.port, master_stdout, master_stderr))
                pyemu.os_utils.run("pestpp-swp {0} /h :{1} 1>{2} 2>{3}". \
                          format(self.pst.filename, self.port, master_stdout, master_stderr))

            except Exception as e:
                self.logger.lraise("error starting condor master: {0}".format(str(e)))
            with open(master_stderr, 'r') as f:
                err_lines = f.readlines()
            if len(err_lines) > 0:
                self.logger.warn("master stderr lines: {0}".
                                 format(','.join([l.strip() for l in err_lines])))

        master_thread = threading.Thread(target=master)
        master_thread.start()
        time.sleep(2.0)
        return master_thread

    def _calc_obs_condor(self,parensemble):
        self.logger.log("evaluating ensemble of size {0} with htcondor".\
                        format(parensemble.shape[0]))

        parensemble.to_csv(self.sweep_in_csv)
        master_thread = self._get_master_thread()
        condor_temp_file = "_condor_submit_stdout.dat"
        condor_err_file = "_condor_submit_stderr.dat"
        self.logger.log("calling condor_submit with submit file {0}".format(self.submit_file))
        try:
            os.system("condor_submit {0} 1>{1} 2>{2}".\
                      format(self.submit_file,condor_temp_file,condor_err_file))
        except Exception as e:
            self.logger.lraise("error in condor_submit: {0}".format(str(e)))
        self.logger.log("calling condor_submit with submit file {0}".format(self.submit_file))
        time.sleep(2.0) #some time for condor to submit the job and echo to stdout
        condor_submit_string = "submitted to cluster"
        with open(condor_temp_file,'r') as f:
            lines = f.readlines()
        self.logger.statement("condor_submit stdout: {0}".\
                              format(','.join([l.strip() for l in lines])))
        with open(condor_err_file,'r') as f:
            err_lines = f.readlines()
        if len(err_lines) > 0:
            self.logger.warn("stderr from condor_submit:{0}".\
                             format([l.strip() for l in err_lines]))
        cluster_number = None
        for line in lines:
            if condor_submit_string in line.lower():
                cluster_number = int(float(line.split(condor_submit_string)[-1]))
        if cluster_number is None:
            self.logger.lraise("couldn't find cluster number...")
        self.logger.statement("condor cluster: {0}".format(cluster_number))
        master_thread.join()
        self.logger.statement("condor master thread exited")
        self.logger.log("calling condor_rm on cluster {0}".format(cluster_number))
        os.system("condor_rm cluster {0}".format(cluster_number))
        self.logger.log("calling condor_rm on cluster {0}".format(cluster_number))
        self.logger.log("evaluating ensemble of size {0} with htcondor".\
                        format(parensemble.shape[0]))


    def _calc_obs_local(self,parensemble):
        '''
        propagate the ensemble forward using sweep.
        '''
        self.logger.log("evaluating ensemble of size {0} locally with sweep".\
                        format(parensemble.shape[0]))
        parensemble.to_csv(self.sweep_in_csv)
        if self.num_slaves > 0:
            master_thread = self._get_master_thread()
            pyemu.utils.start_slaves(self.slave_dir,"pestpp-swp",self.pst.filename,
                                     self.num_slaves,slave_root='..',port=self.port)
            master_thread.join()
        else:
            os.system("pestpp-swp {0}".format(self.pst.filename))

        self.logger.log("evaluating ensemble of size {0} locally with sweep".\
                        format(parensemble.shape[0]))


    def update(self,lambda_mults=[1.0],localizer=None,run_subset=None,use_approx=True):
        raise Exception("EnsembleMethod.update() must be implemented by the derived types")





"""module for error variance analysis, using FOSM
assumptions.
"""
from __future__ import print_function, division
import numpy as np
import pandas as pd
from pyemu.la import LinearAnalysis
from pyemu.mat.mat_handler import Matrix, Jco, Cov

class ErrVar(LinearAnalysis):
    """Derived class for error variance analysis.  Supports 2-term and
    3-term error variance analysis.  Inherits from pyemu.LinearAnalysis.

    
    Parameters
    ----------
    **kwargs : dict
        keyword args to pass to the LinearAnalysis base constructor.
    omitted_parameters :  list
        list of parameters to treat as "omitted".  Passing this argument
        activates 3-term error variance analysis.

    omitted_parcov : varies
        argument that can be cast to a parcov for the omitted parameters.
        If None, omitted_parcov will be formed by extracting from the
        LinearAnalsis.parcov attribute.
    omitted_predictions : varies
        argument that can be cast to a "predictions" attribute for
        prediction sensitivity vectors WRT omitted parameters.  If None,
        these vectors will be extracted from the LinearAnalysis.predictions
        attribute
    kl : bool
        flag to perform KL scaling on the jacobian before error variance
        calculations

    Note
    ----
    There are some additional keyword args that can be passed to active
    the 3-term error variance calculation


    """

    def __init__(self,jco,**kwargs):

        self.__need_omitted = False
        if "omitted_parameters" in kwargs.keys():
             self.omitted_par_arg = kwargs["omitted_parameters"]
             kwargs.pop("omitted_parameters")
             self.__need_omitted = True
        else:
             self.omitted_par_arg = None
        if "omitted_parcov" in kwargs.keys():
            self.omitted_parcov_arg = kwargs["omitted_parcov"]
            kwargs.pop("omitted_parcov")
            self.__need_omitted = True
        else:
            self.omitted_parcov_arg = None

        if "omitted_forecasts" in kwargs.keys():
            self.omitted_predictions_arg = kwargs["omitted_forecasts"]
            kwargs.pop("omitted_forecasts")
            self.__need_omitted = True
        elif "omitted_predictions" in kwargs.keys():
            self.omitted_predictions_arg = kwargs["omitted_predictions"]
            kwargs.pop("omitted_predictions")
            self.__need_omitted = True
        else:
            self.omitted_predictions_arg = None

        kl = False
        if "kl" in kwargs.keys():
            kl = bool(kwargs["kl"])
            kwargs.pop("kl")


        self.__qhalfx = None
        self.__R = None
        self.__R_sv = None
        self.__G = None
        self.__G_sv = None
        self.__I_R = None
        self.__I_R_sv = None
        self.__omitted_jco = None
        self.__omitted_parcov = None
        self.__omitted_predictions = None

        # instantiate the parent class
        super(ErrVar, self).__init__(jco, **kwargs)
        if self.__need_omitted:
            self.log("pre-loading omitted components")
            #self._LinearAnalysis__load_jco()
            #self._LinearAnalysis__load_parcov()
            #self._LinearAnalysis__load_obscov()
            #if self.prediction_arg is not None:
            #    self._LinearAnalysis__load_predictions()
            self.__load_omitted_jco()
            self.__load_omitted_parcov()
            if self.prediction_arg is not None:
                self.__load_omitted_predictions()
            self.log("pre-loading omitted components")
        if kl:
            self.log("applying KL scaling")
            self.apply_karhunen_loeve_scaling()
            self.log("applying KL scaling")

        self.valid_terms = ["null","solution", "omitted", "all"]
        self.valid_return_types = ["parameters", "predictions"]

    def __load_omitted_predictions(self):
        """private: set the omitted_predictions attribute
        """
        # if there are no base predictions
        if self.predictions is None:
            raise Exception("ErrVar.__load_omitted_predictions(): " +
                            "no 'included' predictions is None")
        if self.omitted_predictions_arg is None and \
                        self.omitted_par_arg is None:
            raise Exception("ErrVar.__load_omitted_predictions: " +
                            "both omitted args are None")
        # try to set omitted_predictions by
        # extracting from existing predictions
        if self.omitted_predictions_arg is None and \
                        self.omitted_par_arg is not None:
            # check to see if omitted par names are in each predictions
            found = True
            missing_par,missing_pred = None, None
            for par_name in self.omitted_jco.col_names:
                for prediction in self.predictions_iter:
                    if par_name not in prediction.row_names:
                        found = False
                        missing_par = par_name
                        missing_pred = prediction.col_names[0]
                        break
            if found:
                opreds = []
                # need to access the attribute directly,
                # not a view of attribute
                opred_mat = self._LinearAnalysis__predictions.extract(
                            row_names=self.omitted_jco.col_names)
                opreds = [opred_mat.get(col_names=name) for name in self.forecast_names]

                #for prediction in self._LinearAnalysis__predictions:
                #    opred = prediction.extract(self.omitted_jco.col_names)
                #    opreds.append(opred)
                self.__omitted_predictions = opreds
            else:
                raise Exception("ErrVar.__load_omitted_predictions(): " +
                                " omitted parameter " + str(missing_par) +\
                                " not found in prediction vector " +
                                str(missing_pred))
        elif self.omitted_parcov_arg is not None:
            raise NotImplementedError()

    def __load_omitted_parcov(self):
        """private: set the omitted_parcov attribute
        """
        if self.omitted_parcov_arg is None and self.omitted_par_arg is None:
            raise Exception("ErrVar.__load_omitted_parcov: " +
                            "both omitted args are None")
        # try to set omitted_parcov by extracting from base parcov
        if self.omitted_parcov_arg is None and self.omitted_par_arg is not None:
            # check to see if omitted par names are in parcov
            found = True
            for par_name in self.omitted_jco.col_names:
                if par_name not in self.parcov.col_names:
                    found = False
                    break
            if found:
                # need to access attribute directly, not view of attribute
                self.__omitted_parcov = \
                    self._LinearAnalysis__parcov.extract(
                        row_names=self.omitted_jco.col_names)
            else:
                self.logger.warn("ErrVar.__load_omitted_parun: " +
                                 "no omitted parcov arg passed: " +
                        "setting omitted parcov as identity Matrix")
                self.__omitted_parcov = Cov(
                    x=np.ones(self.omitted_jco.shape[1]),
                    names=self.omitted_jco.col_names, isdiagonal=True)
        elif self.omitted_parcov_arg is not None:
            raise NotImplementedError()

    def __load_omitted_jco(self):
        """private: set the omitted jco attribute
        """
        if self.omitted_par_arg is None:
            raise Exception("ErrVar.__load_omitted: omitted_arg is None")
        if isinstance(self.omitted_par_arg,str):
            if self.omitted_par_arg in self.jco.col_names:
                # need to access attribute directly, not view of attribute
                self.__omitted_jco = \
                    self._LinearAnalysis__jco.extract(
                        col_names=self.omitted_par_arg)
            else:
                # must be a filename
                self.__omitted_jco = self.__fromfile(self.omitted_par_arg)
        # if the arg is an already instantiated Matrix (or jco) object
        elif isinstance(self.omitted_par_arg,Jco) or \
                isinstance(self.omitted_par_arg,Matrix):
            self.__omitted_jco = \
                Jco(x=self.omitted_par_arg.newx(),
                          row_names=self.omitted_par_arg.row_names,
                          col_names=self.omitted_par_arg.col_names)
        # if it is a list, then it must be a list
        # of parameter names in self.jco
        elif isinstance(self.omitted_par_arg,list):
            for arg in self.omitted_par_arg:
                if isinstance(arg,str):
                    assert arg in self.jco.col_names,\
                        "ErrVar.__load_omitted_jco: omitted_jco " +\
                        "arg str not in jco par_names: " + str(arg)
            self.__omitted_jco = \
                self._LinearAnalysis__jco.extract(col_names=self.omitted_par_arg)

    # these property decorators help keep from loading potentially
    # unneeded items until they are called
    # returns a view only - cheap, but can be dangerous

    @property
    def omitted_predictions(self):
        """ get the omitted prediction sensitivity vectors (stored as
        a pyemu.Matrix)

        Returns
        -------
        omitted_predictions : pyemu.Matrix
            a matrix of prediction sensitivity vectors (column wise) to
            omitted parameters

        Note
        ----
        returns a reference
        
        if ErrorVariance.__omitted_predictions is not set, then dynamically load the
        attribute before returning

        """
        if self.__omitted_predictions is None:
            self.log("loading omitted_predictions")
            self.__load_omitted_predictions()
            self.log("loading omitted_predictions")
        return self.__omitted_predictions

    @property
    def omitted_jco(self):
        """get the omitted jco

        Returns
        -------
        omitted_jco : pyemu.Jco

        Note
        ----
        returns a reference
        
        if ErrorVariance.__omitted_jco is None,
        then dynamically load the attribute before returning
        
        """
        if self.__omitted_jco is None:
            self.log("loading omitted_jco")
            self.__load_omitted_jco()
            self.log("loading omitted_jco")
        return self.__omitted_jco

    @property
    def omitted_parcov(self):
        """get the omitted prior parameter covariance matrix

        Returns
        -------
        omitted_parcov : pyemu.Cov

        Note
        ----
        returns a reference
        
        If ErrorVariance.__omitted_parcov is None,
        attribute is dynamically loaded
        
        """
        if self.__omitted_parcov is None:
            self.log("loading omitted_parcov")
            self.__load_omitted_parcov()
            self.log("loading omitted_parcov")
        return self.__omitted_parcov

    def get_errvar_dataframe(self, singular_values=None):
        """get a pandas dataframe of error variance results indexed
            on singular value and (prediction name,<errvar term>)

        Parameters
        ----------
        singular_values : list
            singular values to test.  defaults to
            range(0,min(nnz_obs,nadj_par) + 1)

        Returns
        -------
        pandas.DataFrame : pandas.DataFrame
            multi-indexed pandas dataframe
        
        """
        if singular_values is None:
            singular_values = \
                np.arange(0, min(self.pst.nnz_obs, self.pst.npar_adj) + 1)
        if not isinstance(singular_values, list) and \
                not isinstance(singular_values, np.ndarray):
            singular_values = [singular_values]
        results = {}
        for singular_value in singular_values:
            sv_results = self.variance_at(singular_value)
            for key, val in sv_results.items():
                if key not in results.keys():
                    results[key] = []
                results[key].append(val)
        return pd.DataFrame(results, index=singular_values)


    def get_identifiability_dataframe(self,singular_value=None,precondition=False):
        """get the parameter identifiability as a pandas dataframe

        Parameters
        ----------
        singular_value : int
            the singular spectrum truncation point. Defaults to minimum of
            non-zero-weighted observations and adjustable parameters
        precondition : bool
            flag to use the preconditioned hessian (xtqt + sigma_theta^-1).
            Default is False

        Returns
        -------
        pandas.DataFrame : pandas.DataFrame
            A pandas dataframe of the V_1**2 Matrix with the
            identifiability in the column labeled "ident"

        """
        if singular_value is None:
            singular_value = int(min(self.pst.nnz_obs, self.pst.npar_adj))
        #v1_df = self.qhalfx.v[:, :singular_value].to_dataframe() ** 2
        xtqx = self.xtqx
        if precondition:
            xtqx = xtqx + self.parcov.inv
        #v1_df = self.xtqx.v[:, :singular_value].to_dataframe() ** 2
        v1_df = xtqx.v[:, :singular_value].to_dataframe() ** 2
        v1_df["ident"] = v1_df.sum(axis=1)
        return v1_df

    def variance_at(self, singular_value):
        """get the error variance of all three terms at a singluar value

        Parameters
        ----------
        singular_value : int
            singular value to test

        Returns
        -------
        dict : dict
            dictionary of (err var term,prediction_name), standard_deviation pairs

        """
        results = {}
        results.update(self.first_prediction(singular_value))
        results.update(self.second_prediction(singular_value))
        results.update(self.third_prediction(singular_value))
        return results

    def R(self, singular_value):
        """get resolution Matrix (V_1 * V_1^T) at a singular value

        Parameters
        ----------
        singular_value : int
            singular value to calc R at

        Returns
        -------
        R : pyemu.Matrix
            resolution matrix at singular_value
        """
        if self.__R is not None and singular_value == self.__R_sv:
            return self.__R

        elif singular_value > self.jco.ncol:
            self.__R_sv = self.jco.ncol
            return self.parcov.identity
        else:
            self.log("calc R @" + str(singular_value))
            #v1 = self.qhalfx.v[:, :singular_value]
            v1 = self.xtqx.v[:, :singular_value]
            self.__R = v1 * v1.T
            self.__R_sv = singular_value
            self.log("calc R @" + str(singular_value))
            return self.__R

    def I_minus_R(self,singular_value):
        """get I - R at singular value

        Parameters
        ----------
        singular_value : int
            singular value to calc R at

        Returns
        -------
        I - R : pyemu.Matrix
            identity matrix minus resolution matrix at singular_value

        """
        if self.__I_R is not None and singular_value == self.__I_R_sv:
            return self.__I_R
        else:
            if singular_value > self.jco.ncol:
                return self.parcov.zero
            else:
                #v2 = self.qhalfx.v[:, singular_value:]
                v2 = self.xtqx.v[:, singular_value:]
                self.__I_R = v2 * v2.T
                self.__I_R_sv = singular_value
                return self.__I_R

    def G(self, singular_value):
        """get the parameter solution Matrix at a singular value
            V_1 * S_1^(_1) * U_1^T

        Parameters
        ----------
        singular_value : int
            singular value to calc R at

        Returns
        -------
        G : pyemu.Matrix
            parameter solution matrix at singular value

        """
        if self.__G is not None and singular_value == self.__G_sv:
            return self.__G

        if singular_value == 0:
            self.__G_sv = 0
            self.__G = Matrix(
                x=np.zeros((self.jco.ncol,self.jco.nrow)),
                row_names=self.jco.col_names, col_names=self.jco.row_names)
            return self.__G
        mn = min(self.jco.shape)
        try:
            mn = min(self.pst.npar_adj, self.pst.nnz_obs)
        except:
            pass
        if singular_value > mn:
            self.logger.warn(
                "ErrVar.G(): singular_value > min(npar,nobs):" +
                "resetting to min(npar,nobs): " +
                str(min(self.pst.npar_adj, self.pst.nnz_obs)))
            singular_value = min(self.pst.npar_adj, self.pst.nnz_obs)
        self.log("calc G @" + str(singular_value))
        #v1 = self.qhalfx.v[:, :singular_value]
        v1 = self.xtqx.v[:, :singular_value]
        #s1 = ((self.qhalfx.s[:singular_value]) ** 2).inv
        s1 = (self.xtqx.s[:singular_value]).inv
        self.__G = v1 * s1 * v1.T * self.jco.T * self.obscov.inv
        self.__G_sv = singular_value
        self.__G.row_names = self.jco.col_names
        self.__G.col_names = self.jco.row_names
        self.__G.autoalign = True
        self.log("calc G @" + str(singular_value))
        return self.__G

    def first_forecast(self,singular_value):
        """wrapper around ErrVar.first_forecast
        """
        return self.first_prediction(singular_value)

    def first_prediction(self, singular_value):
        """get the null space term (first term) contribution to prediction error variance
            at a singular value.  used to construct error variance dataframe

        Parameters
        ----------
        singular_value : int
            singular value to calc first term at

        Returns
        -------
        dict : dict
            dictionary of ("first",prediction_names),error variance pairs at singular_value

        """
        if not self.predictions:
            raise Exception("ErrVar.first(): no predictions are set")
        if singular_value > self.jco.ncol:
            zero_preds = {}
            for pred in self.predictions_iter:
                zero_preds[("first", pred.col_names[0])] = 0.0
            return zero_preds
        self.log("calc first term parameter @" + str(singular_value))
        first_term = self.I_minus_R(singular_value).T * self.parcov *\
                     self.I_minus_R(singular_value)
        if self.predictions:
            results = {}
            for prediction in self.predictions_iter:
                results[("first",prediction.col_names[0])] = \
                    float((prediction.T * first_term * prediction).x)
            self.log("calc first term parameter @" + str(singular_value))
            return results

    def first_parameter(self, singular_value):
        """get the null space term (first term) contribution to parameter error variance
            at a singular value

        Parameters
        ----------
        singular_value : int
            singular value to calc first term at

        Returns
        -------
        first_term : pyemu.Cov
            first term contribution to parameter error variance

        """
        self.log("calc first term parameter @" + str(singular_value))
        first_term = self.I_minus_R(singular_value) * self.parcov * \
                     self.I_minus_R(singular_value)
        self.log("calc first term parameter @" + str(singular_value))
        return first_term

    def second_forecast(self,singular_value):
        """wrapper around ErrVar.second_prediction
        """
        return self.second_prediction(singular_value)

    def second_prediction(self, singular_value):
        """get the solution space contribution to predictive error variance
            at a singular value (y^t * G * obscov * G^T * y).  Used to construct
            error variance dataframe

        Parameters
        ----------
        singular_value : int
            singular value to calc second term at

        Returns
        -------
        dict : dict
            dictionary of ("second",prediction_names), error variance

        """

        if not self.predictions:
            raise Exception("ErrVar.second(): not predictions are set")
        self.log("calc second term prediction @" + str(singular_value))

        mn = min(self.jco.shape)
        try:
            mn = min(self.pst.npar_adj, self.pst.nnz_obs)
        except:
            pass
        if singular_value > mn:
            inf_pred = {}
            for pred in self.predictions_iter:
                inf_pred[("second",pred.col_names[0])] = 1.0E+35
            return inf_pred
        else:
            second_term = self.G(singular_value) * self.obscov * \
                          self.G(singular_value).T
            results = {}
            for prediction in self.predictions_iter:
                results[("second",prediction.col_names[0])] = \
                    float((prediction.T * second_term * prediction).x)
            self.log("calc second term prediction @" + str(singular_value))
            return results

    def second_parameter(self, singular_value):
        """get the solution space contribution to parameter error variance
             at a singular value (G * obscov * G^T)

        Parameters
        ----------
        singular_value : int
            singular value to calc second term at

        Returns
        -------
        second_parameter : pyemu.Cov
            second term contribution to parameter error variance

        """
        self.log("calc second term parameter @" + str(singular_value))
        result = self.G(singular_value) * self.obscov * self.G(singular_value).T
        self.log("calc second term parameter @" + str(singular_value))
        return result

    def third_forecast(self,singular_value):
        """wrapper around ErrVar.third_prediction
        """
        return self.third_prediction(singular_value)

    def third_prediction(self,singular_value):
        """get the omitted parameter contribution to prediction error variance
         at a singular value. used to construct error variance dataframe

        Parameters
        ----------
        singular_value : int
            singular value to calc third term at

        Returns
        -------
        dict : dict
            dictionary of ("third",prediction_names),error variance
        """
        if not self.predictions:
            raise Exception("ErrVar.third(): not predictions are set")
        if self.__need_omitted is False:
            zero_preds = {}
            for pred in self.predictions_iter:
                zero_preds[("third", pred.col_names[0])] = 0.0
            return zero_preds
        self.log("calc third term prediction @" + str(singular_value))
        mn = min(self.jco.shape)
        try:
            mn = min(self.pst.npar_adj, self.pst.nnz_obs)
        except:
            pass
        if singular_value > mn:
            inf_pred = {}
            for pred in self.predictions_iter:
                inf_pred[("third",pred.col_names[0])] = 1.0E+35
            return inf_pred
        else:
            results = {}
            for prediction,omitted_prediction in \
                    zip(self.predictions_iter, self.omitted_predictions):
                # comes out as row vector, but needs to be a column vector
                p = ((prediction.T * self.G(singular_value) * self.omitted_jco)
                     - omitted_prediction.T).T
                result = float((p.T * self.omitted_parcov * p).x)
                results[("third", prediction.col_names[0])] = result
            self.log("calc third term prediction @" + str(singular_value))
            return results

    def third_parameter(self, singular_value):
        """get the omitted parameter contribution to parameter error variance
             at a singular value (G * omitted_jco * Sigma_(omitted_pars) * omitted_jco^T * G^T)

        Parameters
        ----------
        singular_value : int
            singular value to calc third term at

        Returns
        -------
        third_parameter : pyemu.Cov
            0.0 if need_omitted is False

        """
        if self.__need_omitted is False:
            return 0.0
        self.log("calc third term parameter @" + str(singular_value))
        GZo = self.G(singular_value) * self.omitted_jco
        result = GZo * self.omitted_parcov * GZo.T
        self.log("calc third term parameter @" + str(singular_value))
        return result




import os
import sys
import platform
import shutil
import subprocess as sp
import warnings
import socket
import time
from datetime import datetime
import pandas as pd
from ..pyemu_warnings import PyemuWarning

ext = ''
bin_path = os.path.join("..","bin")
if "linux" in platform.platform().lower():
    bin_path = os.path.join(bin_path,"linux")
elif "darwin" in platform.platform().lower():
    bin_path = os.path.join(bin_path,"mac")
else:
    bin_path = os.path.join(bin_path,"win")
    ext = '.exe'

bin_path = os.path.abspath(bin_path)
os.environ["PATH"] += os.pathsep + bin_path


def _istextfile(filename, blocksize=512):


    """
        Function found from:
        https://eli.thegreenplace.net/2011/10/19/perls-guess-if-file-is-text-or-binary-implemented-in-python
        Returns True if file is most likely a text file
        Returns False if file is most likely a binary file
        Uses heuristics to guess whether the given file is text or binary,
        by reading a single block of bytes from the file.
        If more than 30% of the chars in the block are non-text, or there
        are NUL ('\x00') bytes in the block, assume this is a binary file.
    """

    import sys
    PY3 = sys.version_info[0] == 3

    # A function that takes an integer in the 8-bit range and returns
    # a single-character byte object in py3 / a single-character string
    # in py2.
    #
    int2byte = (lambda x: bytes((x,))) if PY3 else chr

    _text_characters = (
        b''.join(int2byte(i) for i in range(32, 127)) +
        b'\n\r\t\f\b')
    block = open(filename,'rb').read(blocksize)
    if b'\x00' in block:
        # Files with null bytes are binary
        return False
    elif not block:
        # An empty file is considered a valid text file
        return True

    # Use translate's 'deletechars' argument to efficiently remove all
    # occurrences of _text_characters from the block
    nontext = block.translate(None, _text_characters)
    return float(len(nontext)) / len(block) <= 0.30


def remove_readonly(func, path, excinfo):
    """remove readonly dirs, apparently only a windows issue
    add to all rmtree calls: shutil.rmtree(**,onerror=remove_readonly), wk"""
    os.chmod(path, 128) #stat.S_IWRITE==128==normal
    func(path)


def run_sweep(pe,slave_dir,pst_name=None,num_slaves=10,exe_name="pestpp-swp",local=True,
              binary=False,master_dir="master_runsweep",cleanup=True):

    if pst_name is not  None:
        assert os.path.exists(os.path.join(slave_dir,pst_name))
    else:
        # pst_files = [f for f in os.listdir(template_dir) if f.lower().endswith(".pst")]
        # if len(pst_files) > 1:
        #     raise Exception("run_sweep() error: 'pst_name' is None "+
        #                     "but more than one '.pst' file found in 'template_dir'")
        # if len(pst_files) == 0:
        #     raise Exception("run_sweep() error: 'pst_name' is None and"+\
        #                     " no '.pst' files in 'template_dir'")
        # pst_file = pst_files[0]
        pst = pe.pst
        pst.write(os.path.join(slave_dir,"master_runsweep.pst"))
        pst_name = "master_runsweep.pst"


    # todo: add autodetect to pestpp-swp for sweep_in.jcb
    if binary:
        raise NotImplementedError("pestpp-swp doesn't support autodetect for binary yet")
        pe.to_binary(os.path.join(slave_dir,"sweep_in.jcb"))
    if not binary:
        pe.to_csv(os.path.join(slave_dir,"sweep_in.csv"))
    if not local:
        raise NotImplementedError("condor not supported yet")
    else:
        print(os.getenv("PATH"))
        start_slaves(slave_dir,exe_name,pst_name,num_slaves=num_slaves,slave_root=".",
                     master_dir=master_dir)

    out_file = os.path.join(master_dir,"sweep_out.csv")
    assert os.path.exists(out_file)
    df = pd.read_csv(out_file,index_col=0)
    df.columns = df.columns.map(str.lower)
    if cleanup:
        shutil.rmtree(master_dir)

    return df



def run(cmd_str,cwd='.',verbose=False):
    """ an OS agnostic function to execute a command line

    Parameters
    ----------
    cmd_str : str
        the str to execute with os.system()

    cwd : str
        the directory to execute the command in

    verbose : bool
        flag to echo to stdout complete cmd str

    Note
    ----
    uses platform to detect OS and adds .exe suffix or ./ prefix as appropriate

    for Windows, if os.system returns non-zero, raises exception

    Example
    -------
    ``>>>import pyemu``

    ``>>>pyemu.helpers.run("pestpp pest.pst")``

    """
    bwd = os.getcwd()
    os.chdir(cwd)
    try:
        exe_name = cmd_str.split()[0]
        if "window" in platform.platform().lower():
            if not exe_name.lower().endswith("exe"):
                raw = cmd_str.split()
                raw[0] = exe_name + ".exe"
                cmd_str = ' '.join(raw)
        else:
            if exe_name.lower().endswith('exe'):
                raw = cmd_str.split()
                exe_name = exe_name.replace('.exe','')
                raw[0] = exe_name
                cmd_str = '{0} {1} '.format(*raw)
            if os.path.exists(exe_name) and not exe_name.startswith('./'):
                cmd_str = "./" + cmd_str


    except Exception as e:
        os.chdir(bwd)
        raise Exception("run() error preprocessing command line :{0}".format(str(e)))
    if verbose:
        print("run():{0}".format(cmd_str))

    try:
        ret_val = os.system(cmd_str)
    except Exception as e:
        os.chdir(bwd)
        raise Exception("run() raised :{0}".format(str(e)))
    os.chdir(bwd)
    if "window" in platform.platform().lower():
        if ret_val != 0:
            raise Exception("run() returned non-zero")


def start_slaves(slave_dir,exe_rel_path,pst_rel_path,num_slaves=None,slave_root="..",
                 port=4004,rel_path=None,local=True,cleanup=True,master_dir=None,
                 verbose=False,silent_master=False):
    """ start a group of pest(++) slaves on the local machine

    Parameters
    ----------
    slave_dir :  str
        the path to a complete set of input files
    exe_rel_path : str
        the relative path to the pest(++) executable from within the slave_dir
    pst_rel_path : str
        the relative path to the pst file from within the slave_dir
    num_slaves : int
        number of slaves to start. defaults to number of cores
    slave_root : str
        the root to make the new slave directories in
    rel_path: str
        the relative path to where pest(++) should be run from within the
        slave_dir, defaults to the uppermost level of the slave dir
    local: bool
        flag for using "localhost" instead of hostname on slave command line
    cleanup: bool
        flag to remove slave directories once processes exit
    master_dir: str
        name of directory for master instance.  If master_dir
        exists, then it will be removed.  If master_dir is None,
        no master instance will be started
    verbose : bool
        flag to echo useful information to stdout
    silent_master : bool
        flag to pipe master output to devnull.  This is only for
        pestpp Travis testing. Default is False

    Note
    ----
    if all slaves (and optionally master) exit gracefully, then the slave
    dirs will be removed unless cleanup is false

    Example
    -------
    ``>>>import pyemu``

    start 10 slaves using the directory "template" as the base case and
    also start a master instance in a directory "master".

    ``>>>pyemu.helpers.start_slaves("template","pestpp","pest.pst",10,master_dir="master")``

    """

    assert os.path.isdir(slave_dir)
    assert os.path.isdir(slave_root)
    if num_slaves is None:
        num_slaves = mp.cpu_count()
    else:
        num_slaves = int(num_slaves)
    #assert os.path.exists(os.path.join(slave_dir,rel_path,exe_rel_path))
    exe_verf = True

    if rel_path:
        if not os.path.exists(os.path.join(slave_dir,rel_path,exe_rel_path)):
            #print("warning: exe_rel_path not verified...hopefully exe is in the PATH var")
            exe_verf = False
    else:
        if not os.path.exists(os.path.join(slave_dir,exe_rel_path)):
            #print("warning: exe_rel_path not verified...hopefully exe is in the PATH var")
            exe_verf = False
    if rel_path is not None:
        assert os.path.exists(os.path.join(slave_dir,rel_path,pst_rel_path))
    else:
        assert os.path.exists(os.path.join(slave_dir,pst_rel_path))
    if local:
        hostname = "localhost"
    else:
        hostname = socket.gethostname()

    base_dir = os.getcwd()
    port = int(port)

    if os.path.exists(os.path.join(slave_dir,exe_rel_path)):
        if "window" in platform.platform().lower():
            if not exe_rel_path.lower().endswith("exe"):
                exe_rel_path = exe_rel_path + ".exe"
        else:
            if not exe_rel_path.startswith('./'):
                exe_rel_path = "./" + exe_rel_path

    if master_dir is not None:
        if master_dir != '.' and os.path.exists(master_dir):
            try:
                shutil.rmtree(master_dir,onerror=remove_readonly)#, onerror=del_rw)
            except Exception as e:
                raise Exception("unable to remove existing master dir:" + \
                                "{0}\n{1}".format(master_dir,str(e)))
        if master_dir != '.':
            try:
                shutil.copytree(slave_dir,master_dir)
            except Exception as e:
                raise Exception("unable to copy files from base slave dir: " + \
                                "{0} to master dir: {1}\n{2}".\
                                format(slave_dir,master_dir,str(e)))

        args = [exe_rel_path, pst_rel_path, "/h", ":{0}".format(port)]
        if rel_path is not None:
            cwd = os.path.join(master_dir,rel_path)
        else:
            cwd = master_dir
        if verbose:
            print("master:{0} in {1}".format(' '.join(args),cwd))
        stdout=None
        if silent_master:
            stdout = open(os.devnull,'w')
        try:
            os.chdir(cwd)
            master_p = sp.Popen(args,stdout=stdout)#,stdout=sp.PIPE,stderr=sp.PIPE)
            os.chdir(base_dir)
        except Exception as e:
            raise Exception("error starting master instance: {0}".\
                            format(str(e)))
        time.sleep(1.5) # a few cycles to let the master get ready


    tcp_arg = "{0}:{1}".format(hostname,port)
    procs = []
    slave_dirs = []
    for i in range(num_slaves):
        new_slave_dir = os.path.join(slave_root,"slave_{0}".format(i))
        if os.path.exists(new_slave_dir):
            try:
                shutil.rmtree(new_slave_dir,onerror=remove_readonly)#, onerror=del_rw)
            except Exception as e:
                raise Exception("unable to remove existing slave dir:" + \
                                "{0}\n{1}".format(new_slave_dir,str(e)))
        try:
            shutil.copytree(slave_dir,new_slave_dir)
        except Exception as e:
            raise Exception("unable to copy files from slave dir: " + \
                            "{0} to new slave dir: {1}\n{2}".format(slave_dir,new_slave_dir,str(e)))
        try:
            if exe_verf:
                # if rel_path is not None:
                #     exe_path = os.path.join(rel_path,exe_rel_path)
                # else:
                exe_path = exe_rel_path
            else:
                exe_path = exe_rel_path
            args = [exe_path, pst_rel_path, "/h", tcp_arg]
            #print("starting slave in {0} with args: {1}".format(new_slave_dir,args))
            if rel_path is not None:
                cwd = os.path.join(new_slave_dir,rel_path)
            else:
                cwd = new_slave_dir

            os.chdir(cwd)
            if verbose:
                print("slave:{0} in {1}".format(' '.join(args),cwd))
            with open(os.devnull,'w') as f:
                p = sp.Popen(args,stdout=f,stderr=f)
            procs.append(p)
            os.chdir(base_dir)
        except Exception as e:
            raise Exception("error starting slave: {0}".format(str(e)))
        slave_dirs.append(new_slave_dir)

    if master_dir is not None:
        # while True:
        #     line = master_p.stdout.readline()
        #     if line != '':
        #         print(str(line.strip())+'\r',end='')
        #     if master_p.poll() is not None:
        #         print(master_p.stdout.readlines())
        #         break
        if silent_master:
            # this keeps travis from thinking something is wrong...
            while True:
                rv = master_p.poll()
                if master_p.poll() is not None:
                    break
                print(datetime.now(), "still running")
                time.sleep(5)
        else:
            master_p.wait()
            time.sleep(1.5) # a few cycles to let the slaves end gracefully
        # kill any remaining slaves
        for p in procs:
            p.kill()
    # this waits for sweep to finish, but pre/post/model (sub)subprocs may take longer
    for p in procs:
        p.wait()
    if cleanup:
        cleanit=0
        while len(slave_dirs)>0 and cleanit<100000: # arbitrary 100000 limit
            cleanit=cleanit+1
            for d in slave_dirs:
                try:
                    shutil.rmtree(d,onerror=remove_readonly)
                    slave_dirs.pop(slave_dirs.index(d)) #if successfully removed
                except Exception as e:
                    warnings.warn("unable to remove slavr dir{0}:{1}".format(d,str(e)),PyemuWarning)




import os
import sys
import platform
import shutil
import subprocess as sp
import warnings
import socket
import time
from datetime import datetime
import numpy as np
import pandas as pd
from ..pyemu_warnings import PyemuWarning

def smp_to_ins(smp_filename,ins_filename=None,use_generic_names=False,
               gwutils_compliant=False, datetime_format=None,prefix=''):
    """ create an instruction file for an smp file

    Parameters
    ----------
    smp_filename : str
        existing smp file
    ins_filename: str
        instruction file to create.  If None, create
        an instruction file using the smp filename
        with the ".ins" suffix
    use_generic_names : bool
        flag to force observations names to use a generic
        int counter instead of trying to use a datetime str
    gwutils_compliant : bool
        flag to use instruction set that is compliant with the
        pest gw utils (fixed format instructions).  If false,
        use free format (with whitespace) instruction set
    datetime_format : str
        str to pass to datetime.strptime in the smp_to_dataframe() function
    prefix : str
         a prefix to add to the front of the obsnmes.  Default is ''


    Returns
    -------
    df : pandas.DataFrame
        dataframe instance of the smp file with the observation names and
        instruction lines as additional columns

    """
    if ins_filename is None:
        ins_filename = smp_filename+".ins"
    df = smp_to_dataframe(smp_filename,datetime_format=datetime_format)
    df.loc[:,"ins_strings"] = None
    df.loc[:,"observation_names"] = None
    name_groups = df.groupby("name").groups
    for name,idxs in name_groups.items():
        if not use_generic_names and len(name) <= 11:
            onames = df.loc[idxs,"datetime"].apply(lambda x: prefix+name+'_'+x.strftime("%d%m%Y")).values
        else:
            onames = [prefix+name+"_{0:d}".format(i) for i in range(len(idxs))]
        if False in (map(lambda x :len(x) <= 20,onames)):
            long_names = [oname for oname in onames if len(oname) > 20]
            raise Exception("observation names longer than 20 chars:\n{0}".format(str(long_names)))
        if gwutils_compliant:
            ins_strs = ["l1  ({0:s})39:46".format(on) for on in onames]
        else:
            ins_strs = ["l1 w w w  !{0:s}!".format(on) for on in onames]
        df.loc[idxs,"observation_names"] = onames
        df.loc[idxs,"ins_strings"] = ins_strs

    counts = df.observation_names.value_counts()
    dup_sites = [name for name in counts.index if counts[name] > 1]
    if len(dup_sites) > 0:
        raise Exception("duplicate observation names found:{0}"\
                        .format(','.join(dup_sites)))

    with open(ins_filename,'w') as f:
        f.write("pif ~\n")
        [f.write(ins_str+"\n") for ins_str in df.loc[:,"ins_strings"]]
    return df


def dataframe_to_smp(dataframe,smp_filename,name_col="name",
                     datetime_col="datetime",value_col="value",
                     datetime_format="dd/mm/yyyy",
                     value_format="{0:15.6E}",
                     max_name_len=12):
    """ write a dataframe as an smp file

    Parameters
    ----------
    dataframe : pandas.DataFrame
    smp_filename : str
        smp file to write
    name_col: str
        the column in the dataframe the marks the site namne
    datetime_col: str
        the column in the dataframe that is a datetime instance
    value_col: str
        the column in the dataframe that is the values
    datetime_format: str
        either 'dd/mm/yyyy' or 'mm/dd/yyy'
    value_format: str
        a python float-compatible format

    """
    formatters = {"name":lambda x:"{0:<20s}".format(str(x)[:max_name_len]),
                  "value":lambda x:value_format.format(x)}
    if datetime_format.lower().startswith("d"):
        dt_fmt = "%d/%m/%Y    %H:%M:%S"
    elif datetime_format.lower().startswith("m"):
        dt_fmt = "%m/%d/%Y    %H:%M:%S"
    else:
        raise Exception("unrecognized datetime_format: " +\
                        "{0}".format(str(datetime_format)))

    for col in [name_col,datetime_col,value_col]:
        assert col in dataframe.columns

    dataframe.loc[:,"datetime_str"] = dataframe.loc[:,"datetime"].\
        apply(lambda x:x.strftime(dt_fmt))
    if isinstance(smp_filename,str):
        smp_filename = open(smp_filename,'w')
        # need this to remove the leading space that pandas puts in front
        s = dataframe.loc[:,[name_col,"datetime_str",value_col]].\
                to_string(col_space=0,
                          formatters=formatters,
                          justify=None,
                          header=False,
                          index=False)
        for ss in s.split('\n'):
            smp_filename.write("{0:<s}\n".format(ss.strip()))
    dataframe.pop("datetime_str")


def date_parser(items):
    """ datetime parser to help load smp files

    Parameters
    ----------
    items : iterable
        something or somethings to try to parse into datetimes

    Returns
    -------
    dt : iterable
        the cast datetime things
    """
    try:
        dt = datetime.strptime(items,"%d/%m/%Y %H:%M:%S")
    except Exception as e:
        try:
            dt = datetime.strptime(items,"%m/%d/%Y %H:%M:%S")
        except Exception as ee:
            raise Exception("error parsing datetime string" +\
                            " {0}: \n{1}\n{2}".format(str(items),str(e),str(ee)))
    return dt


def smp_to_dataframe(smp_filename,datetime_format=None):
    """ load an smp file into a pandas dataframe (stacked in wide format)

    Parameters
    ----------
    smp_filename : str
        smp filename to load
    datetime_format : str
        should be either "%m/%d/%Y %H:%M:%S" or "%d/%m/%Y %H:%M:%S"
        If None, then we will try to deduce the format for you, which
        always dangerous

    Returns
    -------
    df : pandas.DataFrame

    """

    if datetime_format is not None:
        date_func = lambda x: datetime.strptime(x,datetime_format)
    else:
        date_func = date_parser
    df = pd.read_csv(smp_filename, delim_whitespace=True,
                     parse_dates={"datetime":["date","time"]},
                     header=None,names=["name","date","time","value"],
                     dtype={"name":object,"value":np.float64},
                     na_values=["dry"],
                     date_parser=date_func)
    return df



"""module for logging pyemu progress
"""
from __future__ import print_function, division
from datetime import datetime
import warnings
from .pyemu_warnings import PyemuWarning
import copy

class Logger(object):
    """ a basic class for logging events during the linear analysis calculations
        if filename is passed, then a file handle is opened.

    Parameters
    ----------
    filename : str
        Filename to write logged events to. If False, no file will be created,
        and logged events will be displayed on standard out.
    echo : bool
        Flag to cause logged events to be echoed to the screen.

    Attributes
    ----------
    items : dict
        Dictionary holding events to be logged.  If a log entry is
        not in `items`, then it is treated as a new entry with the string
        being the key and the datetime as the value.  If a log entry is
        in `items`, then the end time and delta time are written and
        the item is popped from the keys.

    """
    def __init__(self,filename, echo=False):
        self.items = {}
        self.echo = bool(echo)
        if filename == True:
            self.echo = True
            self.filename = None
        elif filename:
            self.filename = filename
            self.f = open(filename, 'w')
            self.t = datetime.now()
            self.log("opening " + str(filename) + " for logging")
        else:
            self.filename = None


    def statement(self,phrase):
        """ log a one time statement

        Parameters
        ----------
        phrase : str
            statement to log

        """
        t = datetime.now()
        s = str(t) + ' ' + str(phrase) + '\n'
        if self.echo:
            print(s,end='')
        if self.filename:
            self.f.write(s)
            self.f.flush()


    def log(self,phrase):
        """log something that happened.  The first time phrase is passed the
        start time is saved.  The second time the phrase is logged, the
        elapsed time is written

        Parameters
        ----------
            phrase : str
                the thing that happened
        """
        pass
        t = datetime.now()
        if phrase in self.items.keys():
            s = str(t) + ' finished: ' + str(phrase) + " took: " + \
                str(t - self.items[phrase]) + '\n'
            if self.echo:
                print(s,end='')
            if self.filename:
                self.f.write(s)
                self.f.flush()
            self.items.pop(phrase)
        else:
            s = str(t) + ' starting: ' + str(phrase) + '\n'
            if self.echo:
                print(s,end='')
            if self.filename:
                self.f.write(s)
                self.f.flush()
            self.items[phrase] = copy.deepcopy(t)

    def warn(self,message):
        """write a warning to the log file.

        Parameters
        ----------
        message : str
            the warning text
        """
        s = str(datetime.now()) + " WARNING: " + message + '\n'
        if self.echo:
            print(s,end='')
        if self.filename:
            self.f.write(s)
            self.f.flush
        warnings.warn(s,PyemuWarning)

    def lraise(self,message):
        """log an exception, close the log file, then raise the exception

        Parameters
        ----------
        message : str
            the exception message

        Raises
        ------
            exception with message
        """
        s = str(datetime.now()) + " ERROR: " + message + '\n'
        print(s,end='')
        if self.filename:
            self.f.write(s)
            self.f.flush
            self.f.close()
        raise Exception(message)



"""This module contains helpers and default values that support
the pyemu.Pst object.
"""
from __future__ import print_function, division
import os, sys
import stat
import warnings
from datetime import datetime
import numpy as np
import pandas as pd
pd.options.display.max_colwidth = 100

import pyemu
from ..pyemu_warnings import PyemuWarning
#formatters
#SFMT = lambda x: "{0:>20s}".format(str(x.decode()))
def SFMT(item):
    try:
        s = "{0:<20s} ".format(item.decode())
    except:
        s = "{0:<20s} ".format(str(item))
    return s

SFMT_LONG = lambda x: "{0:<50s} ".format(str(x))
IFMT = lambda x: "{0:<10d} ".format(int(x))
FFMT = lambda x: "{0:<20.10E} ".format(float(x))

def str_con(item):
    if len(item) == 0:
        return np.NaN
    return item.lower().strip()

pst_config = {}

# parameter stuff
pst_config["tied_dtype"] = np.dtype([("parnme", "U20"), ("partied","U20")])
pst_config["tied_fieldnames"] = ["parnme","partied"]
pst_config["tied_format"] = {"parnme":SFMT,"partied":SFMT}
pst_config["tied_converters"] = {"parnme":str_con,"partied":str_con}
pst_config["tied_defaults"] = {"parnme":"dum","partied":"dum"}

pst_config["par_dtype"] = np.dtype([("parnme", "U20"), ("partrans","U20"),
                                   ("parchglim","U20"),("parval1", np.float64),
                                   ("parlbnd",np.float64),("parubnd",np.float64),
                                   ("pargp","U20"),("scale", np.float64),
                                   ("offset", np.float64),("dercom",np.int)])
pst_config["par_fieldnames"] = "PARNME PARTRANS PARCHGLIM PARVAL1 PARLBND PARUBND " +\
                              "PARGP SCALE OFFSET DERCOM"
pst_config["par_fieldnames"] = pst_config["par_fieldnames"].lower().strip().split()
pst_config["par_format"] = {"parnme": SFMT, "partrans": SFMT,
                           "parchglim": SFMT, "parval1": FFMT,
                           "parlbnd": FFMT, "parubnd": FFMT,
                           "pargp": SFMT, "scale": FFMT,
                           "offset": FFMT, "dercom": IFMT}
pst_config["par_converters"] = {"parnme": str_con, "pargp": str_con,
                                "parval1":np.float,"parubnd":np.float,
                                "parlbnd":np.float,"scale":np.float,
                                "offset":np.float}
pst_config["par_defaults"] = {"parnme":"dum","partrans":"log","parchglim":"factor",
                             "parval1":1.0,"parlbnd":1.1e-10,"parubnd":1.1e+10,
                             "pargp":"pargp","scale":1.0,"offset":0.0,"dercom":1}


# parameter group stuff
pst_config["pargp_dtype"] = np.dtype([("pargpnme", "U20"), ("inctyp","U20"),
                                   ("derinc", np.float64),
                                   ("derinclb",np.float64),("forcen","U20"),
                                   ("derincmul",np.float64),("dermthd", "U20"),
                                   ("splitthresh", np.float64),("splitreldiff",np.float64),
                                      ("splitaction","U20")])
pst_config["pargp_fieldnames"] = "PARGPNME INCTYP DERINC DERINCLB FORCEN DERINCMUL " +\
                        "DERMTHD SPLITTHRESH SPLITRELDIFF SPLITACTION"
pst_config["pargp_fieldnames"] = pst_config["pargp_fieldnames"].lower().strip().split()

pst_config["pargp_format"] = {"pargpnme":SFMT,"inctyp":SFMT,"derinc":FFMT,"forcen":SFMT,
                      "derincmul":FFMT,"dermthd":SFMT,"splitthresh":FFMT,
                      "splitreldiff":FFMT,"splitaction":SFMT}

pst_config["pargp_converters"] = {"pargpnme":str_con,"inctyp":str_con,
                         "dermethd":str_con,"derinc":np.float,"derinclb":np.float,
                         "splitaction":str_con,"forcen":str_con,"derincmul":np.float}
pst_config["pargp_defaults"] = {"pargpnme":"pargp","inctyp":"relative","derinc":0.01,
                       "derinclb":0.0,"forcen":"switch","derincmul":2.0,
                     "dermthd":"parabolic","splitthresh":1.0e-5,
                       "splitreldiff":0.5,"splitaction":"smaller"}


# observation stuff
pst_config["obs_fieldnames"] = "OBSNME OBSVAL WEIGHT OBGNME".lower().split()
pst_config["obs_dtype"] = np.dtype([("obsnme","U20"),("obsval",np.float64),
                           ("weight",np.float64),("obgnme","U20")])
pst_config["obs_format"] = {"obsnme": SFMT, "obsval": FFMT,
                   "weight": FFMT, "obgnme": SFMT}
pst_config["obs_converters"] = {"obsnme": str_con, "obgnme": str_con,
                                "weight":np.float,"obsval":np.float}
pst_config["obs_defaults"] = {"obsnme":"dum","obsval":1.0e+10,
                     "weight":1.0,"obgnme":"obgnme"}


# prior info stuff
pst_config["null_prior"] = pd.DataFrame({"pilbl": None,
                                    "obgnme": None}, index=[])
pst_config["prior_format"] = {"pilbl": SFMT, "equation": SFMT_LONG,
                     "weight": FFMT, "obgnme": SFMT}
pst_config["prior_fieldnames"] = ["pilbl","equation", "weight", "obgnme"]


# other containers
pst_config["model_command"] = []
pst_config["template_files"] = []
pst_config["input_files"] = []
pst_config["instruction_files"] = []
pst_config["output_files"] = []
pst_config["other_lines"] = []
pst_config["tied_lines"] = []
pst_config["regul_lines"] = []
pst_config["pestpp_options"] = {}


def read_resfile(resfile):
        """load a residual file into a pandas.DataFrame

        Parameters
        ----------
        resfile : str
            residual file name

        Returns
        -------
        pandas.DataFrame : pandas.DataFrame

        """
        assert os.path.exists(resfile),"read_resfile() error: resfile " +\
                                       "{0} not found".format(resfile)
        converters = {"name": str_con, "group": str_con}
        f = open(resfile, 'r')
        while True:
            line = f.readline()
            if line == '':
                raise Exception("Pst.get_residuals: EOF before finding "+
                                "header in resfile: " + resfile)
            if "name" in line.lower():
                header = line.lower().strip().split()
                break
        res_df = pd.read_csv(f, header=None, names=header, sep="\s+",
                                 converters=converters)
        res_df.index = res_df.name
        f.close()
        return res_df

def res_from_en(pst,enfile):
    """load ensemble file for residual into a pandas.DataFrame

        Parameters
        ----------
        enfile : str
            ensemble file name

        Returns
        -------
        pandas.DataFrame : pandas.DataFrame

        """
    converters = {"name": str_con, "group": str_con}
    try: #substitute ensemble for res, 'base' if there, otherwise mean
        obs=pst.observation_data
        if isinstance(enfile,str):
            df=pd.read_csv(enfile,converters=converters)
            df.columns=df.columns.str.lower()
            df = df.set_index('real_name').T.rename_axis('name').rename_axis(None, 1)
        else:
            df = enfile.T
        if 'base' in df.columns:
            df['modelled']=df['base']
            df['std']=df.std(axis=1)
        else:
            df['modelled']=df.mean(axis=1)
            df['std']=df.std(axis=1)
        #probably a more pandastic way to do this
        res_df=df[['modelled','std']].copy()
        res_df['group']=obs.loc[:,'obgnme'].copy()
        res_df['measured']=obs['obsval'].copy()
        res_df['weight']=obs['weight'].copy()
        res_df['residual']=res_df['measured']-res_df['modelled']
    except Exception as e:
        raise Exception("Pst.res_from_en:{0}".format(str(e)))
    return res_df

def read_parfile(parfile):
    """load a pest-compatible .par file into a pandas.DataFrame

    Parameters
    ----------
    parfile : str
        pest parameter file name

    Returns
    -------
    pandas.DataFrame : pandas.DataFrame

    """
    assert os.path.exists(parfile), "Pst.parrep(): parfile not found: " +\
                                    str(parfile)
    f = open(parfile, 'r')
    header = f.readline()
    par_df = pd.read_csv(f, header=None,
                             names=["parnme", "parval1", "scale", "offset"],
                             sep="\s+")
    par_df.index = par_df.parnme
    return par_df

def write_parfile(df,parfile):
    """ write a pest parameter file from a dataframe

    Parameters
    ----------
    df : (pandas.DataFrame)
        dataframe with column names that correspond to the entries
        in the parameter data section of a pest control file
    parfile : str
        name of the parameter file to write

    """
    columns = ["parnme","parval1","scale","offset"]
    formatters = {"parnme":lambda x:"{0:20s}".format(x),
                  "parval1":lambda x:"{0:20.7E}".format(x),
                  "scale":lambda x:"{0:20.7E}".format(x),
                  "offset":lambda x:"{0:20.7E}".format(x)}

    for col in columns:
        assert col in df.columns,"write_parfile() error: " +\
                                 "{0} not found in df".format(col)
    with open(parfile,'w') as f:
        f.write("single point\n")
        f.write(df.to_string(col_space=0,
                      columns=columns,
                      formatters=formatters,
                      justify="right",
                      header=False,
                      index=False,
                      index_names=False) + '\n')

def parse_tpl_file(tpl_file):
    """ parse a pest template file to get the parameter names

    Parameters
    ----------
    tpl_file : str
        template file name

    Returns
    -------
    par_names : list
        list of parameter names

    """
    par_names = set()
    with open(tpl_file,'r') as f:
        try:
            header = f.readline().strip().split()
            assert header[0].lower() in ["ptf","jtf"],\
                "template file error: must start with [ptf,jtf], not:" +\
                str(header[0])
            assert len(header) == 2,\
                "template file error: header line must have two entries: " +\
                str(header)

            marker = header[1]
            assert len(marker) == 1,\
                "template file error: marker must be a single character, not:" +\
                str(marker)
            for line in f:
                par_line = set(line.lower().strip().split(marker)[1::2])
                par_names.update(par_line)
                #par_names.extend(par_line)
                #for p in par_line:
                #    if p not in par_names:
                #        par_names.append(p)
        except Exception as e:
            raise Exception("error processing template file " +\
                            tpl_file+" :\n" + str(e))
    #par_names = [pn.strip().lower() for pn in par_names]
    #seen = set()
    #seen_add = seen.add
    #return [x for x in par_names if not (x in seen or seen_add(x))]
    return [p.strip() for p in list(par_names)]


def write_input_files(pst):
    """write parameter values to a model input files using a template files with
    current parameter values (stored in Pst.parameter_data.parval1).
    This is a simple implementation of what PEST does.  It does not
    handle all the special cases, just a basic function...user beware

    Parameters
    ----------
    pst : (pyemu.Pst)
        a Pst instance

    """
    par = pst.parameter_data
    par.loc[:,"parval1_trans"] = (par.parval1 * par.scale) + par.offset
    for tpl_file,in_file in zip(pst.template_files,pst.input_files):
        write_to_template(pst.parameter_data.parval1_trans,tpl_file,in_file)

def write_to_template(parvals,tpl_file,in_file):
    """ write parameter values to model input files using template files

    Parameters
    ----------
    parvals : dict or pandas.Series
        a way to look up parameter values using parameter names
    tpl_file : str
        template file
    in_file : str
        input file

    """
    f_in = open(in_file,'w')
    f_tpl = open(tpl_file,'r')
    header = f_tpl.readline().strip().split()
    assert header[0].lower() in ["ptf", "jtf"], \
        "template file error: must start with [ptf,jtf], not:" + \
        str(header[0])
    assert len(header) == 2, \
        "template file error: header line must have two entries: " + \
        str(header)

    marker = header[1]
    assert len(marker) == 1, \
        "template file error: marker must be a single character, not:" + \
        str(marker)
    for line in f_tpl:
        if marker not in line:
            f_in.write(line)
        else:
            line = line.rstrip()
            par_names = line.lower().split(marker)[1::2]
            par_names = [name.strip() for name in par_names]
            start,end = get_marker_indices(marker,line)
            assert len(par_names) == len(start)
            new_line = line[:start[0]]
            between = [line[e:s] for s,e in zip(start[1:],end[:-1])]
            for i,name in enumerate(par_names):
                s,e = start[i],end[i]
                w = e - s
                if w > 15:
                    d = 6
                else:
                    d = 3
                fmt = "{0:" + str(w)+"."+str(d)+"E}"
                val_str = fmt.format(parvals[name])
                new_line += val_str
                if i != len(par_names) - 1:
                    new_line += between[i]
            new_line += line[end[-1]:]
            f_in.write(new_line+'\n')
    f_tpl.close()
    f_in.close()


def get_marker_indices(marker,line):
    """ method to find the start and end parameter markers
    on a template file line.  Used by write_to_template()

    Parameters
    ----------
    marker : str
        template file marker char
    line : str
        template file line

    Returns
    -------
    indices : list
        list of start and end indices (zero based)

    """
    indices = [i for i, ltr in enumerate(line) if ltr == marker]
    start = indices[0:-1:2]
    end = [i+1 for i in indices[1::2]]
    assert len(start) == len(end)
    return start,end


def parse_ins_file(ins_file):
    """parse a pest instruction file to get observation names

    Parameters
    ----------
    ins_file : str
        instruction file name

    Returns
    -------
    list of observation names

    """

    obs_names = []
    with open(ins_file,'r') as f:
        header = f.readline().strip().split()
        assert header[0].lower() in ["pif","jif"],\
            "instruction file error: must start with [pif,jif], not:" +\
            str(header[0])
        marker = header[1]
        assert len(marker) == 1,\
            "instruction file error: marker must be a single character, not:" +\
            str(marker)
        for line in f:
            line = line.lower()
            if marker in line:
                raw = line.lower().strip().split(marker)
                for item in raw[::2]:
                    obs_names.extend(parse_ins_string(item))
            else:
                obs_names.extend(parse_ins_string(line.strip()))
    #obs_names = [on.strip().lower() for on in obs_names]
    return obs_names


def parse_ins_string(string):
    """ split up an instruction file line to get the observation names

    Parameters
    ----------
    string : str
        instruction file line

    Returns
    -------
    obs_names : list
        list of observation names

    """
    istart_markers = ["[","(","!"]
    iend_markers = ["]",")","!"]

    obs_names = []

    idx = 0
    while True:
        if idx >= len(string) - 1:
            break
        char = string[idx]
        if char in istart_markers:
            em = iend_markers[istart_markers.index(char)]
            # print("\n",idx)
            # print(string)
            # print(string[idx+1:])
            # print(string[idx+1:].index(em))
            # print(string[idx+1:].index(em)+idx+1)
            eidx = min(len(string),string[idx+1:].index(em)+idx+1)
            obs_name = string[idx+1:eidx]
            if obs_name.lower() != "dum":
                obs_names.append(obs_name)
            idx = eidx + 1
        else:
            idx += 1
    return obs_names


def populate_dataframe(index,columns, default_dict, dtype):
    """ helper function to populate a generic Pst dataframe attribute.  This
    function is called as part of constructing a generic Pst instance

    Parameters
    ----------
    index : (varies)
        something to use as the dataframe index
    columns: (varies)
        something to use as the dataframe columns
    default_dict : (dict)
        dictionary of default values for columns
    dtype : numpy.dtype
        dtype used to cast dataframe columns

    Returns
    -------
    new_df : pandas.DataFrame

    """
    new_df = pd.DataFrame(index=index,columns=columns)
    for fieldname,dt in zip(columns,dtype.descr):
        default = default_dict[fieldname]
        new_df.loc[:,fieldname] = default
        new_df.loc[:,fieldname] = new_df.loc[:,fieldname].astype(dt[1])
    return new_df


def generic_pst(par_names=["par1"],obs_names=["obs1"],addreg=False):
    """generate a generic pst instance.  This can used to later fill in
    the Pst parts programatically.

    Parameters
    ----------
    par_names : (list)
        parameter names to setup
    obs_names : (list)
        observation names to setup

    Returns
    -------
    new_pst : pyemu.Pst

    """
    if not isinstance(par_names,list):
        par_names = list(par_names)
    if not isinstance(obs_names,list):
        obs_names = list(obs_names)
    new_pst = pyemu.Pst("pest.pst",load=False)
    pargp_data = populate_dataframe(["pargp"], new_pst.pargp_fieldnames,
                                    new_pst.pargp_defaults, new_pst.pargp_dtype)
    new_pst.parameter_groups = pargp_data

    par_data = populate_dataframe(par_names,new_pst.par_fieldnames,
                                  new_pst.par_defaults,new_pst.par_dtype)
    par_data.loc[:,"parnme"] = par_names
    par_data.index = par_names
    par_data.sort_index(inplace=True)
    new_pst.parameter_data = par_data
    obs_data = populate_dataframe(obs_names,new_pst.obs_fieldnames,
                                  new_pst.obs_defaults,new_pst.obs_dtype)
    obs_data.loc[:,"obsnme"] = obs_names
    obs_data.index = obs_names
    obs_data.sort_index(inplace=True)
    new_pst.observation_data = obs_data

    new_pst.template_files = ["file.tpl"]
    new_pst.input_files = ["file.in"]
    new_pst.instruction_files = ["file.ins"]
    new_pst.output_files = ["file.out"]
    new_pst.model_command = ["model.bat"]

    new_pst.prior_information = new_pst.null_prior

    #new_pst.other_lines = ["* singular value decomposition\n","1\n",
    #                       "{0:d} {1:15.6E}\n".format(new_pst.npar_adj,1.0E-6),
    #                       "1 1 1\n"]
    if addreg:
        new_pst.zero_order_tikhonov()

    return new_pst


def pst_from_io_files(tpl_files,in_files,ins_files,out_files,pst_filename=None):
    """ generate a new pyemu.Pst instance from model interface files.  This
    function is emulated in the Pst.from_io_files() class method.

    Parameters
    ----------
    tpl_files : (list)
        template file names
    in_files : (list)
        model input file names
    ins_files : (list)
        instruction file names
    out_files : (list)
        model output file names
    pst_filename : str
        filename to save new pyemu.Pst.  If None, Pst is not written.
        default is None

    Returns
    -------
    new_pst : pyemu.Pst

    """

    warnings.warn("pst_from_io_files has moved to pyemu.helpers and is also "+\
                  "now avaiable as a Pst class method (Pst.from_io_files())",PyemuWarning)
    from pyemu import helpers
    return helpers.pst_from_io_files(tpl_files=tpl_files,in_files=in_files,
                              ins_files=ins_files,out_files=out_files,
                              pst_filename=pst_filename)


def try_run_inschek(pst):
    """ attempt to run INSCHEK for each instruction file, model output
    file pair in a pyemu.Pst.  If the run is successful, the INSCHEK written
    .obf file is used to populate the pst.observation_data.obsval attribute

    Parameters
    ----------
    pst : (pyemu.Pst)

    """
    for ins_file,out_file in zip(pst.instruction_files,pst.output_files):
        df = _try_run_inschek(ins_file,out_file)
        if df is not None:
            pst.observation_data.loc[df.index, "obsval"] = df.obsval


def try_process_ins_file(ins_file,out_file=None):
    assert os.path.exists(ins_file),"instruction file {0} not found".format(ins_file)
    try:
        obs_names = parse_ins_file(ins_file)
    except Exception as e:
        raise Exception("error parsing ins file {0} for obs names: {1}".format(ins_file,str(e)))

    if out_file is None:
        out_file = ins_file.replace(".ins","")
    if not os.path.exists(out_file):
        print("out file {0} not found".format(out_file))
        return pd.DataFrame({"obsnme":obs_names},index="obsnme")
    try:
        f_ins = open(ins_file,'r')
        f_out = open(out_file,'r')

        header = f_ins.readline().lower().strip().split()
        assert header[0] == "pif"
        marker = header[1]
        icount,ocount = 1,0
        obsnme,obsval = [],[]
        for iline in f_ins:
            iraw = iline.lower().strip().split()
            if iraw[0][0] != 'l':
                raise Exception("not support instruction:{0}".format(iraw[0]))
            num_lines = int(iraw[0][1:])
            for i in range(num_lines):
                oline = f_out.readline().lower().strip()
                ocount += 1
            if '(' in oline:
                raise Exception("semi-fixed obs index instructions not supported")
            elif '[' in oline:
                raise Exception("fixed obs index instructions not supported")
            elif marker in oline:
                raise Exception("marker-based file seeking not supported")
            oraw = oline.strip().split()
            if oline[0] in [' ','   ']:
                oraw.insert(0,'')
            oc = 0
            for ir in iraw[1:]:
                if ir == 'w':
                    oc += 1
                    if oc > len(oraw):
                        raise Exception("out file line {0} too short".format(ocount))
                elif '!' in ir:
                    n = ir.replace('!','')
                    try:
                        v = float(oraw[oc])
                    except Exception as e:
                        raise Exception("error processing ins {0} for obs {1}, string: {2} on line {3} (ins line {4}):{5}".
                                        format(ir,n,oline,ocount,icount,str(e)))
                    obsnme.append(n)
                    obsval.append(v)
                    oc += 1

        f_ins.close()
        f_out.close()
        df = pd.DataFrame({"obsnme":obsnme,"obsval":obsval},index=obsnme)

        return df
    except Exception as e:
        print("error processing ins file {0}: {1}".format(ins_file,str(e)))
        return pd.DataFrame({"obsnme":obs_names},index=obs_names)



def _try_run_inschek(ins_file,out_file,cwd='.'):

    try:
        pyemu.os_utils.run("inschek {0} {1}".format(ins_file, out_file),cwd=cwd)
        obf_file = os.path.join(cwd,ins_file.replace(".ins", ".obf"))
        df = pd.read_csv(obf_file, delim_whitespace=True,
                         skiprows=0, index_col=0, names=["obsval"])
        df.index = df.index.map(str.lower)
        return df
    except Exception as e:
        print("error using inschek for instruction file {0}:{1}".
              format(ins_file, str(e)))
        print("observations in this instruction file will have" +
              "generic values.")
        return None


def get_phi_comps_from_recfile(recfile):
    """read the phi components from a record file by iteration

    Parameters
    ----------
    recfile : str
        pest record file name

    Returns
    -------
    iters : dict
        nested dictionary of iteration number, {group,contribution}

    """
    iiter = 1
    iters = {}
    f = open(recfile,'r')
    while True:
        line = f.readline()
        if line == '':
            break
        if "starting phi for this iteration" in line.lower() or \
            "final phi" in line.lower():
            contributions = {}
            while True:
                line = f.readline()
                if line == '':
                    break
                if "contribution to phi" not in line.lower():
                    iters[iiter] = contributions
                    iiter += 1
                    break
                raw = line.strip().split()
                val = float(raw[-1])
                group = raw[-3].lower().replace('\"', '')
                contributions[group] = val
    return iters


def del_rw(action, name, exc):
    os.chmod(name, stat.S_IWRITE)
    os.remove(name)

def start_slaves(slave_dir,exe_rel_path,pst_rel_path,num_slaves=None,slave_root="..",
                 port=4004,rel_path=None):


    warnings.warn("deprecation warning:start_slaves() has moved to the utils.helpers module",PyemuWarning)
    from pyemu.utils import start_slaves
    start_slaves(slave_dir,exe_rel_path,pst_rel_path,num_slaves=num_slaves,slave_root=slave_root,
                 port=port,rel_path=rel_path)

def res_from_obseravtion_data(observation_data):
    """create a generic residual dataframe filled with np.NaN for
    missing information

    Parameters
    ----------
    observation_data : pandas.DataFrame
        pyemu.Pst.observation_data

    Returns
    -------
    res_df : pandas.DataFrame

    """
    res_df = observation_data.copy()
    res_df.loc[:, "name"] = res_df.pop("obsnme")
    res_df.loc[:, "measured"] = res_df.pop("obsval")
    res_df.loc[:, "group"] = res_df.pop("obgnme")
    res_df.loc[:, "modelled"] = np.NaN
    res_df.loc[:, "residual"] = np.NaN
    return res_df

def clean_missing_exponent(pst_filename,clean_filename="clean.pst"):
    """fixes the issue where some terrible fortran program may have
    written a floating point format without the 'e' - like 1.0-3, really?!

    Parameters
    ----------
    pst_filename : str
        the pest control file
    clean_filename : str
        the new pest control file to write. Default is "clean.pst"

    Returns
    -------
    None


    """
    lines = []
    with open(pst_filename,'r') as f:
        for line in f:
            line = line.lower().strip()
            if '+' in line:
                raw = line.split('+')
                for i,r in enumerate(raw[:-1]):
                    if r[-1] != 'e':
                        r = r + 'e'
                    raw[i] = r
                lines.append('+'.join(raw))
            else:
                lines.append(line)
    with open(clean_filename,'w') as f:
        for line in lines:
            f.write(line+'\n')

def csv_to_ins_file(csv_filename,ins_filename=None,only_cols=None,only_rows=None,
                    marker='~',includes_header=True,includes_index=True,prefix=''):

    # process the csv_filename in case it is a dataframe
    if isinstance(csv_filename,str):
        df = pd.read_csv(csv_filename,index_col=0)
        df.columns = df.columns.map(str.lower)
        df.index = df.index.map(lambda x: str(x).lower())
    else:
        df = csv_filename

    # process only_cols
    if only_cols is None:
        only_cols = set(df.columns.map(str.lower))
    else:
        if isinstance(only_cols,str): # incase it is a single name
            only_cols = [only_cols]
        only_cols = set(only_cols)

    if only_rows is None:
        only_rows = set(df.index.map(str.lower))
    else:
        if isinstance(only_rows,str): # incase it is a single name
            only_rows = [only_rows]
        only_rows = set(only_rows)

    # process the row labels, handling duplicates
    rlabels = []
    row_visit = {}
    only_rlabels = []
    for rname in df.index:
        rname = str(rname).strip().lower()

        if rname in row_visit:
            rsuffix = str(int(row_visit[rname] + 1))
            row_visit[rname] += 1
        else:
            row_visit[rname] = 1
            rsuffix = ''
        rlabel = rname + rsuffix
        rlabels.append(rlabel)
        if rname in only_rows:
            only_rlabels.append(rlabel)
    only_rlabels = set(only_rlabels)

    #process the col labels, handling duplicates
    clabels = []
    col_visit = {}
    only_clabels = []
    for cname in df.columns:
        cname = str(cname).strip().lower()
        if cname in col_visit:
            csuffix = str(int(col_visit[cname]+1))
            col_visit[cname] += 1
        else:
            col_visit[cname] = 1
            csuffix = ''
        clabel = cname + csuffix
        clabels.append(clabel)
        if cname in only_cols:
            only_clabels.append(clabel)

    if ins_filename is None:
        if not isinstance(csv_filename,str):
            raise Exception("ins_filename is None but csv_filename is not string")
        ins_filename = csv_filename + ".ins"
    row_visit, col_visit = {},{}
    onames = []
    ovals = []
    with open(ins_filename,'w') as f:
        f.write("pif ~\nl1\n")
        for i,rlabel in enumerate(rlabels):
            if includes_header:
                f.write("l1 ") #skip the row (index) label
            for j,clabel in enumerate(clabels):
                if rlabel in only_rlabels and clabel in only_clabels:
                    oname = prefix+rlabel+"_"+clabel
                    onames.append(oname)
                    ovals.append(df.iloc[i,j])
                else:
                    oname = "dum"
                if j == 0:
                    if includes_index:
                        f.write(" {0},{0} ".format(marker))
                else:
                    f.write(" {0},{0} ".format(marker))
                f.write(" !{0}! ".format(oname))
            f.write('\n')
    odf = pd.DataFrame({"obsnme":onames,"obsval":ovals},index=onames)
    return odf













